import tempfile
from copy import deepcopy
from pathlib import Path

import numpy as np
from google.protobuf import json_format
from pandas.testing import assert_frame_equal

import locan.constants
from locan import FileType
from locan.locan_io.locdata import manifest_pb2
from locan.locan_io.locdata.smlm_io import (
    _change_upper_to_lower_keys,
    load_SMLM_file,
    load_SMLM_header,
    load_SMLM_manifest,
    manifest_file_info_from_locdata,
    manifest_format_from_locdata,
    manifest_from_locdata,
    save_SMLM,
)


def test__change_upper_to_lower_keys():
    assert "text" == _change_upper_to_lower_keys("TEXT")


def test_manifest_format_from_locdata(locdata_2d):
    format = manifest_format_from_locdata(locdata_2d)
    assert isinstance(format, manifest_pb2.Format)
    # print(format)


def test_manifest_file_info_from_locdata(locdata_2d):
    file_info = manifest_file_info_from_locdata(locdata_2d)
    assert isinstance(file_info, manifest_pb2.FileInfo)
    # print(file_info)


def test_manifest_from_locdata(locdata_2d):
    manifest = manifest_from_locdata(locdata_2d)
    assert isinstance(manifest, manifest_pb2.Manifest)
    # print(manifest)

    # check for capital letters in manifest that should be introduced by manifest_pb2.Manifest
    json_string = json_format.MessageToJson(
        manifest,
        preserving_proto_field_name=True,
        always_print_fields_with_no_presence=False,
    )
    assert json_string
    assert "BINARY" in json_string
    assert "INT" in json_string

    # check that capital letters are eliminated by _change_upper_to_lower_keys in manifest_from_locdata
    manifest = manifest_from_locdata(locdata_2d, return_json_string=True)
    assert isinstance(manifest, str)
    assert "binary" in manifest
    assert "BINARY" not in manifest
    assert "int" in manifest
    assert "INT" not in manifest
    # print(manifest)


def test_get_correct_column_names_from_SMLM_header():
    columns = load_SMLM_header(
        path=locan.ROOT_DIR / "tests/test_data/SMLM_dstorm_data.smlm"
    )
    assert columns == [
        "original_index",
        "position_x",
        "local_background",
        "chi_square",
        "intensity",
        "frame",
        "position_y",
    ]


def test_load_SMLM_manifest():
    manifest = load_SMLM_manifest(
        path=locan.ROOT_DIR / "tests/test_data/SMLM_dstorm_data.smlm"
    )
    for key in ["format_version", "formats", "files"]:
        assert key in manifest.keys()


def test_loading_SMLM_file():
    file_path = locan.ROOT_DIR / "tests/test_data/SMLM_dstorm_data.smlm"
    locdata = load_SMLM_file(path=file_path, nrows=10)
    assert np.array_equal(
        locdata.data.columns,
        [
            "original_index",
            "position_x",
            "local_background",
            "chi_square",
            "intensity",
            "frame",
            "position_y",
        ],
    )
    assert len(locdata) == 10
    # print(locdata.data)
    assert locdata.meta.file.type == FileType.SMLM.value
    assert locdata.meta.file.path == str(file_path)


def test_save_and_load_smlm(locdata_2d):
    locdata = deepcopy(locdata_2d)
    # introduce different dtypes in data
    locdata.data["position_x"] = locdata.data["position_x"].astype("float64")
    locdata.data["position_y"] = locdata.data["position_y"].astype("int32")
    locdata.data["frame"] = locdata.data["frame"].astype("int32")
    locdata.data["intensity"] = locdata.data["intensity"].astype("int32")
    assert all(locdata.data.dtypes.values == ["float64", "int32", "int32", "int32"])
    # print(locdata.data)

    with tempfile.TemporaryDirectory() as tmp_directory:
        # save smlm file
        file_path = Path(tmp_directory) / "locdata.smlm"
        save_SMLM(locdata, path=file_path)

        # read back smlm manifest
        manifest = load_SMLM_manifest(path=file_path)
        for key in ["format_version", "formats", "files"]:
            assert key in manifest.keys()

        # read back smlm file
        locdata_ = load_SMLM_file(path=file_path, convert=False)
        assert len(locdata_) == len(locdata)
        assert locdata_.meta.identifier != locdata.meta.identifier
        assert locdata_.properties.keys() == locdata.properties.keys()

        # print(locdata.data)
        # print(locdata.data.dtypes)
        # print(manifest_from_locdata(locdata))
        assert_frame_equal(locdata_.data, locdata.data, check_dtype=True)

        # passing manifest
        manifest = manifest_from_locdata(locdata, return_json_string=True)
        save_SMLM(locdata, path=file_path, manifest=manifest)

        manifest = manifest_from_locdata(locdata)
        save_SMLM(locdata, path=file_path, manifest=manifest)

        # read back smlm manifest
        manifest = load_SMLM_manifest(path=file_path)
        for key in ["format_version", "formats", "files"]:
            assert key in manifest.keys()

        # read back smlm file
        locdata_ = load_SMLM_file(path=file_path)
        assert_frame_equal(locdata_.data, locdata.data, check_dtype=False)



"""
Process localization data.

This module contains procedures to process localization data to yield new
datasets.
All functions typically take LocData objects as input and return LocData
properties or whole new LocData objects.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   aggregate
   cluster
   filter
   register
   properties
   tracking
   transform
"""

from __future__ import annotations

from importlib import import_module

# explicit re-exports for nested submodules are required by myp
from locan.process.aggregate import *
from locan.process.cluster import (
    cluster_by_bin as cluster_by_bin,
    cluster_dbscan as cluster_dbscan,
    cluster_hdbscan as cluster_hdbscan,
    serial_clustering as serial_clustering,
)
from locan.process.filter import *
from locan.process.register import *
from locan.process.properties import (
    distance_to_region as distance_to_region,
    distance_to_region_boundary as distance_to_region_boundary,
    inertia_moments as inertia_moments,
    max_distance as max_distance,
    range_from_collection as range_from_collection,
    ranges as ranges,
    statistics as statistics,
)
from locan.process.tracking import *
from locan.process.transform import (
    bunwarp as bunwarp,
    overlay as overlay,
    standardize as standardize,
    transform_affine as transform_affine,
    transform_counts_to_photons as transform_counts_to_photons,
)

submodules: list[str] = [
    "aggregate",
    "cluster",
    "filter",
    "register",
    "properties",
    "tracking",
    "transform",
]

__all__: list[str] = []

for submodule in submodules:
    module_ = import_module(name=f".{submodule}", package="locan.process")
    __all__.extend(module_.__all__)



"""
Functions to compute locdata properties.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, NamedTuple

import numpy as np
import numpy.typing as npt
from scipy.spatial.distance import pdist
from shapely.geometry import Point

from locan.data.regions.region import Region2D, RoiRegion

if TYPE_CHECKING:
    from locan.data.locdata import LocData
    from locan.data.regions.region import Region

__all__: list[str] = [
    "distance_to_region",
    "distance_to_region_boundary",
    "max_distance",
    "inertia_moments",
]

logger = logging.getLogger(__name__)


class InertiaMoments(NamedTuple):
    eigenvalues: npt.NDArray[Any]
    eigenvectors: npt.NDArray[Any]
    variance_explained: list[float]
    orientation: float
    eccentricity: float


def distance_to_region(locdata: LocData, region: Region) -> npt.NDArray[np.float64]:
    """
    Determine the distance to the nearest point within `region` for all
    localizations.
    Returns zero if localization is within the region.

    Parameters
    ----------
    locdata
        Localizations for which distances are determined.

    region
        Region from which the closest point is selected.

    Returns
    --------
    npt.NDArray[np.float64]
        Distance for each localization.
    """
    distances = np.full(len(locdata), 0.0)
    if isinstance(region, (Region2D, RoiRegion)):
        for i, point in enumerate(locdata.coordinates):
            distances[i] = Point(point).distance(region.shapely_object)
    else:
        raise NotImplementedError("Region must be Region2D object.")

    return distances


def distance_to_region_boundary(
    locdata: LocData, region: Region
) -> npt.NDArray[np.float64]:
    """
    Determine the distance to the nearest region boundary for all
    localizations.
    Returns a positive value regardless of weather the point is within or
    outside the region.

    Parameters
    ----------
    locdata
        Localizations for which distances are determined.

    region
        Region from which the closest point is selected.

    Returns
    --------
    npt.NDArray[np.float64]
        Distance for each localization.
    """
    distances = np.full(len(locdata), 0.0)
    if isinstance(region, (Region2D, RoiRegion)):
        for i, point in enumerate(locdata.coordinates):
            distances[i] = Point(point).distance(region.shapely_object.boundary)
    else:
        raise NotImplementedError("Region must be Region2D object.")

    return distances


def max_distance(locdata: LocData) -> dict[str, float]:
    """
    Return maximum of all distances between any two localizations in locdata.

    Parameters
    ----------
    locdata
        Localization data

    Returns
    -------
    dict[str, float]
        A dict with key `max_distance` and the corresponding value being the
        maximum distance.
    """
    if not locdata or len(locdata) == 1:
        points = None
    elif len(locdata) == 2:
        points = locdata.coordinates
    elif locdata.convex_hull is None:
        points = None
    else:
        points = locdata.convex_hull.vertices

    distances = np.nan if points is None else pdist(points)
    distance = float(np.nanmax(distances))
    return {"max_distance": distance}


def inertia_moments(points: npt.ArrayLike) -> InertiaMoments:
    """
    Return inertia moments (or principal components) and related properties
    for the given points.
    Inertia moments are represented by eigenvalues (and corresponding
    eigenvectors) of the covariance matrix.
    Variance_explained represents the eigenvalues normalized to the sum of all
    eigenvalues.
    For 2-dimensional data, orientation is the angle (in degrees) between the
    principal axis with the largest variance and the x-axis.
    Also, for 2-dimensional data, eccentricity is computed as
    e=Sqrt(1-M_min/M_max).

    Parameters
    ----------
    points
        Coordinates of input points with shape (npoints, ndim).

    Returns
    -------
    InertiaMoments
        A tuple with eigenvalues, eigenvectors, variance_explained,
        orientation, eccentricity

    Note
    ----
    The data is not standardized.
    """
    points = np.asarray(points)
    covariance_matrix = np.cov(points.T)
    eigen_values, eigen_vectors = np.linalg.eig(covariance_matrix)
    variance_explained = [
        eigen_value / sum(eigen_values) for eigen_value in eigen_values
    ]
    index_max_eigen_value = np.argmax(eigen_values)

    if np.shape(points)[1] == 2:
        orientation = np.degrees(
            np.arctan2(
                eigen_vectors[1][index_max_eigen_value],
                eigen_vectors[0][index_max_eigen_value],
            )
        )
        eccentricity = np.sqrt(1 - np.min(eigen_values) / np.max(eigen_values))
    else:  # todo implement for 3d
        logger.warning(
            "Orientation and eccentricity have not yet been implemented for 3D."
        )
        orientation = np.nan
        eccentricity = np.nan

    inertia_moments_ = InertiaMoments(
        eigen_values, eigen_vectors, variance_explained, orientation, eccentricity
    )
    return inertia_moments_



"""
Utility to visualize test data.

Show all test datasets that are available in locan/tests/test_data.
"""

from __future__ import annotations

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import tifffile as tif

import locan as lc
from locan import ROOT_DIR

PATH_TEST_DATA = ROOT_DIR / "tests/test_data"


def get_files(path: Path = ROOT_DIR / "tests/test_data") -> list[Path]:
    files = path.rglob(pattern="*.*")
    file_list = [file.relative_to(path) for file in files]
    return file_list


def show_decode_dstorm_data() -> None:
    file = "decode_dstorm_data.h5"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_decode_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y", "position_z"]].describe())
    lc.render_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        bin_size=0.1,
    )
    lc.scatter_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        index=False,
    )
    plt.show()


def show_decode_dstorm_data_empty() -> None:
    file = "decode_dstorm_data_empty.h5"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_decode_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    lc.render_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        bin_size=0.1,
    )
    lc.scatter_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        index=False,
    )
    plt.show()


def show_Elyra_dstorm_data() -> None:
    file = "Elyra_dstorm_data.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_Elyra_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y"]].describe())
    lc.render_2d_mpl(locdata=locdata, bin_size=100)
    plt.show()


def show_five_blobs() -> None:
    file = "five_blobs.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_txt_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y"]].describe())
    lc.render_2d_mpl(locdata=locdata, bin_size=10)
    lc.scatter_2d_mpl(locdata=locdata, index=False)
    plt.show()


def show_five_blobs_3D() -> None:
    import napari

    file = "five_blobs_3D.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_txt_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    viewer = lc.render_3d_napari(
        locdata=locdata,
        bin_size=10,
        cmap="viridis",
        gamma=0.1,
    )
    viewer.add_points(locdata.coordinates)
    napari.run()


def show_images() -> None:
    import napari

    file = "images.tif"
    file_path = PATH_TEST_DATA / file
    image_stack = tif.imread(str(file_path))

    print("\n" + file + "\n")
    print("shape:", np.shape(image_stack))  # type: ignore

    viewer = napari.Viewer()
    viewer.add_image(image_stack, name="images")
    napari.run()


def show_Nanoimager_dstorm_data() -> None:
    file = "Nanoimager_dstorm_data.csv"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_Nanoimager_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        bin_size=100,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


def show_npc_gp210_data() -> None:
    file = "npc_gp210.asdf"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_asdf_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata.projection(coordinate_labels=["position_x", "position_y"]),
        bin_size=10,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


def show_protobuf_message_metadata_pb2_Metadata_v0p11() -> None:
    file = "protobuf_message_metadata_pb2.Metadata_v0p11"
    file_path = PATH_TEST_DATA / file
    metadata_new = lc.data.metadata_pb2.Metadata()
    with open(file_path, "rb") as file_:
        metadata_new.ParseFromString(file_.read())
    print("\n" + file + "\n")
    print(metadata_new)


def show_protobuf_message_metadata_pb2_Metadata_v0p12() -> None:
    file = "protobuf_message_metadata_pb2.Metadata_v0p12"
    file_path = PATH_TEST_DATA / file
    metadata_new = lc.data.metadata_pb2.Metadata()
    with open(file_path, "rb") as file_:
        metadata_new.ParseFromString(file_.read())
    print("\n" + file + "\n")
    print(metadata_new)


def show_rapidSTORM_dstorm_data() -> None:
    file = "rapidSTORM_dstorm_data.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_rapidSTORM_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=100,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


def show_rapidSTORM_dstorm_track_data() -> None:
    file = "rapidSTORM_dstorm_track_data.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_rapidSTORM_track_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=100,
    )
    lc.scatter_2d_mpl(locdata=locdata, index=False)
    plt.show()


def show_rapidSTORM_from_images() -> None:
    file = "rapidSTORM_from_images.txt"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_rapidSTORM_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=100,
        rescale=lc.Trafo.EQUALIZE,
    )
    lc.scatter_2d_mpl(locdata=locdata, index=False)
    plt.show()


def show_smap_dstorm_data() -> None:
    import napari

    file = "smap_dstorm_data.mat"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_SMAP_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y", "position_z"]].describe())
    lc.render_3d_napari(
        locdata=locdata,
        bin_size=100,
        cmap="viridis",
    )
    napari.run()


def show_SMLM_dstorm_data() -> None:
    file = "SMLM_dstorm_data.smlm"
    file_path = PATH_TEST_DATA / file
    locdata: lc.LocData = lc.load_SMLM_file(path=file_path)  # type: ignore
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=100,
    )
    lc.scatter_2d_mpl(locdata=locdata, index=False)
    plt.show()


def show_Thunderstorm_dstorm_data() -> None:
    file = "Thunderstorm_dstorm_data.csv"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_thunderstorm_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y"]].describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=100,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


def show_transform_BunwarpJ_transformation_elastic_green() -> None:
    file = "transform/BunwarpJ_transformation_elastic_green.txt"
    file_path = PATH_TEST_DATA / file
    with open(file_path, "rb") as file_:
        data = file_.read()
    print("\n" + file + "\n")
    print(data)


def show_transform_BunwarpJ_transformation_raw_green() -> None:
    file = "transform/BunwarpJ_transformation_raw_green.txt"
    file_path = PATH_TEST_DATA / file
    with open(file_path, "rb") as file_:
        data = file_.read()
    print("\n" + file + "\n")
    print(data)


def show_transform_rapidSTORM_beads_green() -> None:
    file = "transform/rapidSTORM_beads_green.asdf"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_asdf_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y"]].describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=1,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


def show_transform_rapidSTORM_beads_red() -> None:
    file = "transform/rapidSTORM_beads_red.asdf"
    file_path = PATH_TEST_DATA / file
    locdata = lc.load_asdf_file(path=file_path)
    print("\n" + file + "\n")
    print(locdata.meta)
    print(list(locdata.data.columns), "\n")
    print(locdata.data.loc[:, ["position_x", "position_y"]].describe())
    lc.render_2d_mpl(
        locdata=locdata,
        bin_size=1,
        rescale=lc.Trafo.EQUALIZE,
    )
    plt.show()


if __name__ == "__main__":
    for f in get_files():
        print(f)
    # show_decode_dstorm_data()
    # show_decode_dstorm_data_empty()
    # show_Elyra_dstorm_data()
    # show_five_blobs()
    # show_five_blobs_3D()
    # show_images()
    # show_Nanoimager_dstorm_data()
    # show_npc_gp210_data()
    # show_protobuf_message_metadata_pb2_Metadata_v0p11()
    # show_protobuf_message_metadata_pb2_Metadata_v0p12()
    # show_rapidSTORM_dstorm_data()
    # show_rapidSTORM_dstorm_track_data()
    # show_rapidSTORM_from_images()
    # show_smap_dstorm_data()
    # show_SMLM_dstorm_data()
    # show_Thunderstorm_dstorm_data()
    # show_transform_BunwarpJ_transformation_elastic_green()
    # show_transform_BunwarpJ_transformation_raw_green()
    # show_transform_rapidSTORM_beads_green()
    # show_transform_rapidSTORM_beads_red()



import matplotlib.pyplot as plt  # this import is needed for visual inspection
import numpy as np
import pandas as pd
import pytest

from locan import LocData, NearestNeighborDistances, simulate_uniform
from locan.analysis.nearest_neighbor import (
    NNDistances_csr_2d,
    NNDistances_csr_3d,
    _DistributionFits,
    pdf_nnDistances_csr_2D,
    pdf_nnDistances_csr_3D,
)

# fixtures


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def other_locdata_simple():
    locdata_dict = {
        "position_x": [10, 11],
        "position_y": [10, 11],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


# tests


def test_pdf_nnDistances_csr_2D():
    pdf_values = pdf_nnDistances_csr_2D(np.array([0, 1, 2]), density=1)
    assert pdf_values == pytest.approx([0, 2.71521056e-001, 4.38232365e-05])


def test_pdf_nnDistances_csr_3D():
    pdf_values = pdf_nnDistances_csr_3D(np.array([0, 1, 2]), density=1)
    assert pdf_values == pytest.approx([0, 1.90564233e-01, 1.40579528e-13])


def test_NNDistances_csr_2d():
    dist = NNDistances_csr_2d()
    assert dist.shapes == "density"
    assert dist.pdf(x=np.array([0, 1, 2]), density=1) == pytest.approx(
        [0, 2.71521056e-01, 4.38232365e-05]
    )


def test_NNDistances_csr_3d():
    dist = NNDistances_csr_3d()
    assert dist.shapes == "density"
    assert dist.pdf(x=np.array([0, 1, 2]), density=1) == pytest.approx(
        [0, 1.90564233e-01, 1.40579528e-13]
    )


def test_Nearest_neighbor_distances_empty(caplog):
    nn_1 = NearestNeighborDistances().compute(LocData())
    nn_1.hist()
    assert caplog.record_tuples == [
        ("locan.analysis.nearest_neighbor", 30, "Locdata is empty.")
    ]


def test_DistributionFits(locdata_simple):
    nn_1 = NearestNeighborDistances()
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    nn_1.compute(locdata_simple)
    assert nn_1.dimension == 2
    assert nn_1.localization_density == 0.25
    # print(nn_1.results)
    assert len(nn_1.results) == 5
    assert all(nn_1.results.columns == ["nn_distance", "nn_index"])
    assert nn_1.results["nn_index"].iloc[0] == 1
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert ds.parameter_dict() == {
        "density": pytest.approx(0.058984374999999166),
        "loc": 0,
        "scale": 1,
    }
    ds.fit(with_constraints=False)
    assert ds.parameter_dict() == pytest.approx(
        {
            "density": 0.8900132393760514,
            "loc": 0.35972770464810505,
            "scale": 3.348388757318209,
        }
    )
    # ds.plot()

    nn_1.hist(fit=True)
    assert nn_1.distribution_statistics.parameter_dict() == {
        "density": pytest.approx(0.058984374999999166),
        "loc": 0,
        "scale": 1,
    }
    # plt.show()

    plt.close("all")


def test_DistributionFits_k2(locdata_simple):
    nn_1 = NearestNeighborDistances(k=2)
    assert repr(nn_1) == "NearestNeighborDistances(k=2)"
    nn_1.compute(locdata_simple)
    assert nn_1.dimension == 2
    assert nn_1.localization_density == 0.25
    # print(nn_1.results)
    assert len(nn_1.results) == 5
    assert all(nn_1.results.columns == ["nn_distance", "nn_index"])
    assert nn_1.results["nn_index"].iloc[0] == 2
    assert nn_1.distribution_statistics is None
    nn_1.fit_distributions()
    assert nn_1.distribution_statistics.parameter_dict() == {
        "density": pytest.approx(0.028906249999999137),
        "loc": 0,
        "scale": 1,
    }

    nn_1.hist(fit=True)
    # plt.show()

    plt.close("all")


def test_Nearest_neighbor_distances(locdata_simple, other_locdata_simple):
    nn_2 = NearestNeighborDistances().compute(
        locdata_simple, other_locdata=other_locdata_simple
    )
    # print(nn_2.results)
    assert nn_2.results["nn_distance"].iloc[0] == pytest.approx(14.142135623730951)


def test_NearestNeighborDistances_3d(locdata_3d):
    # print(locdata_3d.data)
    nn_1 = NearestNeighborDistances().compute(locdata_3d)
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    assert nn_1.dimension == 3
    assert nn_1.localization_density == 0.075
    # print(nn_1.results)
    assert len(nn_1.results) == 6
    assert all(nn_1.results.columns == ["nn_distance", "nn_index"])
    assert nn_1.results["nn_index"].iloc[0] == 4
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert ds.parameter_dict() == {
        "density": pytest.approx(0.007666015624999118),
        "loc": 0,
        "scale": 1,
    }
    ds.fit(with_constraints=False)
    assert ds.parameter_dict() == pytest.approx(
        {
            "density": 1.0047585928889682,
            "loc": 2.4220565244971968,
            "scale": 1.3178818824051617,
        }
    )
    # ds.plot()

    nn_1.hist(fit=True)
    assert nn_1.distribution_statistics.parameter_dict() == {
        "density": pytest.approx(0.007666015624999118),
        "loc": 0,
        "scale": 1,
    }
    # plt.show()

    plt.close("all")


def test_NearestNeighborDistances_1d(caplog, locdata_1d):
    nn_1 = NearestNeighborDistances().compute(locdata_1d)
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    assert nn_1.dimension == 1
    assert nn_1.localization_density == 1.5
    # print(nn_1.results)
    assert len(nn_1.results) == 6
    assert all(nn_1.results.columns == ["nn_distance", "nn_index"])
    assert nn_1.results["nn_index"].iloc[0] == 1
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert caplog.record_tuples == [
        (
            "locan.analysis.nearest_neighbor",
            30,
            "No fit model for 1 dimensions available.",
        )
    ]
    assert ds.parameter_dict() == {}
    ds.plot()

    nn_1.hist(fit=True)
    # plt.show()

    plt.close("all")


@pytest.mark.visual
def test_NearestNeighborDistances_2d_random():
    locdata = simulate_uniform(n_samples=10_000, region=((0, 1), (0, 1)))
    nn_1 = NearestNeighborDistances().compute(locdata)
    assert nn_1.localization_density == locdata.properties["localization_density_bb"]
    assert len(nn_1.results) == 10_000
    nn_1.hist(fit=True)
    plt.show()


@pytest.mark.visual
def test_NearestNeighborDistances_3d_random():
    locdata = simulate_uniform(n_samples=10_000, region=((0, 1), (0, 1), (0, 1)))
    nn_1 = NearestNeighborDistances().compute(locdata)
    assert nn_1.localization_density == locdata.properties["localization_density_bb"]
    assert len(nn_1.results) == 10_000
    nn_1.hist(fit=True)
    plt.show()



"""
Standard analysis procedures.

This module contains classes and functions for carrying out standardized
analysis procedures on localization data.
All functions typically take LocData objects as input and provide an
analysis class with derived analysis results
and standard functions for presentation.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   accumulation_analysis
   blinking
   cbc
   convex_hull_expectation
   drift
   grouped_property_expectation
   localization_precision
   localization_property
   localization_property_2d
   localization_property_correlations
   localizations_per_frame
   nearest_neighbor
   pipeline
   position_variance_expectation
   ripley
   subpixel_bias
   uncertainty
"""

from __future__ import annotations

from importlib import import_module

from locan.analysis.accumulation_analysis import *
from locan.analysis.blinking import *
from locan.analysis.cbc import *
from locan.analysis.convex_hull_expectation import *
from locan.analysis.drift import *
from locan.analysis.grouped_property_expectation import *
from locan.analysis.localization_precision import *
from locan.analysis.localization_property import *
from locan.analysis.localization_property_2d import *
from locan.analysis.localization_property_correlations import *
from locan.analysis.localizations_per_frame import *
from locan.analysis.nearest_neighbor import *
from locan.analysis.pipeline import *
from locan.analysis.position_variance_expectation import *
from locan.analysis.ripley import *
from locan.analysis.subpixel_bias import *
from locan.analysis.uncertainty import *

submodules: list[str] = [
    "accumulation_analysis",
    "blinking",
    "cbc",
    "convex_hull_expectation",
    "drift",
    "grouped_property_expectation",
    "localization_precision",
    "localization_property",
    "localization_property_2d",
    "localization_property_correlations",
    "localizations_per_frame",
    "nearest_neighbor",
    "pipeline",
    "position_variance_expectation",
    "ripley",
    "subpixel_bias",
    "uncertainty",
]

__all__: list[str] = []

for submodule in submodules:
    module_ = import_module(name=f".{submodule}", package="locan.analysis")
    __all__.extend(module_.__all__)



from . import (
    analysis,
    configuration,
    constants,
    data,
    datasets,
    dependencies,
    gui,
    locan_io,
    process,
    rois,
    scripts,
    simulation,
    tests,
    utils,
    visualize,
)

# the following explicit re-exports are required by mypy
from .analysis import (
    AccumulationClusterCheck as AccumulationClusterCheck,
    BlinkStatistics as BlinkStatistics,
    ConvexHullExpectation as ConvexHullExpectation,
    ConvexHullExpectationBatch as ConvexHullExpectationBatch,
    CoordinateBasedColocalization as CoordinateBasedColocalization,
    Drift as Drift,
    DriftComponent as DriftComponent,
    GroupedPropertyExpectation as GroupedPropertyExpectation,
    LocalizationPrecision as LocalizationPrecision,
    LocalizationProperty as LocalizationProperty,
    LocalizationProperty2d as LocalizationProperty2d,
    LocalizationPropertyCorrelations as LocalizationPropertyCorrelations,
    LocalizationUncertainty as LocalizationUncertainty,
    LocalizationsPerFrame as LocalizationsPerFrame,
    NearestNeighborDistances as NearestNeighborDistances,
    Pipeline as Pipeline,
    PositionVarianceExpectation as PositionVarianceExpectation,
    RipleysHFunction as RipleysHFunction,
    RipleysKFunction as RipleysKFunction,
    RipleysLFunction as RipleysLFunction,
    SubpixelBias as SubpixelBias,
)
from .configuration import (
    COLORMAP_DEFAULTS as COLORMAP_DEFAULTS,
    DATASETS_DIR as DATASETS_DIR,
    N_JOBS as N_JOBS,
    QT_BINDING as QT_BINDING,
    RENDER_ENGINE as RENDER_ENGINE,
    TQDM_DISABLE as TQDM_DISABLE,
    TQDM_LEAVE as TQDM_LEAVE,
)
from .constants import (
    DECODE_KEYS as DECODE_KEYS,
    ELYRA_KEYS as ELYRA_KEYS,
    FileType as FileType,
    HullType as HullType,
    NANOIMAGER_KEYS as NANOIMAGER_KEYS,
    PROPERTY_KEYS as PROPERTY_KEYS,
    PropertyDescription as PropertyDescription,
    PropertyKey as PropertyKey,
    RAPIDSTORM_KEYS as RAPIDSTORM_KEYS,
    ROOT_DIR as ROOT_DIR,
    RenderEngine as RenderEngine,
    SMAP_KEYS as SMAP_KEYS,
    SMLM_KEYS as SMLM_KEYS,
    THUNDERSTORM_KEYS as THUNDERSTORM_KEYS,
)
from .data import (
    AlphaComplex as AlphaComplex,
    AlphaShape as AlphaShape,
    AxisOrientedCuboid as AxisOrientedCuboid,
    AxisOrientedHypercuboid as AxisOrientedHypercuboid,
    BoundingBox as BoundingBox,
    ConvexHull as ConvexHull,
    Cuboid as Cuboid,
    Ellipse as Ellipse,
    EmptyRegion as EmptyRegion,
    Interval as Interval,
    LocData as LocData,
    MultiPolygon as MultiPolygon,
    OrientedBoundingBox as OrientedBoundingBox,
    Polygon as Polygon,
    Rectangle as Rectangle,
    Region as Region,
    Region1D as Region1D,
    Region2D as Region2D,
    Region3D as Region3D,
    RegionND as RegionND,
    RoiRegion as RoiRegion,
    expand_region as expand_region,
    load_metadata_from_toml as load_metadata_from_toml,
    merge_metadata as merge_metadata,
    message_scheme as message_scheme,
    metadata_from_toml_string as metadata_from_toml_string,
    metadata_to_formatted_string as metadata_to_formatted_string,
    regions_union as regions_union,
    surrounding_region as surrounding_region,
)
from .datasets import load_npc as load_npc, load_tubulin as load_tubulin
from .dependencies import (
    EXTRAS_REQUIRE as EXTRAS_REQUIRE,
    HAS_DEPENDENCY as HAS_DEPENDENCY,
    IMPORT_NAMES as IMPORT_NAMES,
    INSTALL_REQUIRES as INSTALL_REQUIRES,
    QtBindings as QtBindings,
    needs_package as needs_package,
    needs_package_version as needs_package_version,
)
from .gui import (
    file_dialog as file_dialog,
    set_file_path_dialog as set_file_path_dialog,
)
from .locan_io import (
    Files as Files,
    convert_property_names as convert_property_names,
    convert_property_types as convert_property_types,
    find_file_upstream as find_file_upstream,
    load_Elyra_file as load_Elyra_file,
    load_Elyra_header as load_Elyra_header,
    load_Nanoimager_file as load_Nanoimager_file,
    load_Nanoimager_header as load_Nanoimager_header,
    load_SMAP_file as load_SMAP_file,
    load_SMAP_header as load_SMAP_header,
    load_SMLM_file as load_SMLM_file,
    load_SMLM_header as load_SMLM_header,
    load_SMLM_manifest as load_SMLM_manifest,
    load_asdf_file as load_asdf_file,
    load_decode_file as load_decode_file,
    load_decode_header as load_decode_header,
    load_locdata as load_locdata,
    load_rapidSTORM_file as load_rapidSTORM_file,
    load_rapidSTORM_header as load_rapidSTORM_header,
    load_rapidSTORM_track_file as load_rapidSTORM_track_file,
    load_rapidSTORM_track_header as load_rapidSTORM_track_header,
    load_thunderstorm_file as load_thunderstorm_file,
    load_thunderstorm_header as load_thunderstorm_header,
    load_txt_file as load_txt_file,
    manifest_file_info_from_locdata as manifest_file_info_from_locdata,
    manifest_format_from_locdata as manifest_format_from_locdata,
    manifest_from_locdata as manifest_from_locdata,
    save_SMAP_csv as save_SMAP_csv,
    save_SMLM as save_SMLM,
    save_asdf as save_asdf,
    save_thunderstorm_csv as save_thunderstorm_csv,
)
from .process import (
    Bins as Bins,
    Selector as Selector,
    bunwarp as bunwarp,
    cluster_by_bin as cluster_by_bin,
    cluster_dbscan as cluster_dbscan,
    cluster_hdbscan as cluster_hdbscan,
    distance_to_region as distance_to_region,
    distance_to_region_boundary as distance_to_region_boundary,
    exclude_sparse_points as exclude_sparse_points,
    filter_condition as filter_condition,
    histogram as histogram,
    inertia_moments as inertia_moments,
    link_locdata as link_locdata,
    localizations_in_cluster_regions as localizations_in_cluster_regions,
    max_distance as max_distance,
    overlay as overlay,
    random_subset as random_subset,
    range_from_collection as range_from_collection,
    ranges as ranges,
    register_cc as register_cc,
    register_icp as register_icp,
    select_by_condition as select_by_condition,
    select_by_image_mask as select_by_image_mask,
    select_by_region as select_by_region,
    serial_clustering as serial_clustering,
    standardize as standardize,
    statistics as statistics,
    track as track,
    transform_affine as transform_affine,
    transform_counts_to_photons as transform_counts_to_photons,
)
from .rois import Roi as Roi, rasterize as rasterize
from .simulation import (
    add_drift as add_drift,
    make_Matern as make_Matern,
    make_NeymanScott as make_NeymanScott,
    make_Poisson as make_Poisson,
    make_Thomas as make_Thomas,
    make_cluster as make_cluster,
    make_dstorm as make_dstorm,
    make_uniform as make_uniform,
    randomize as randomize,
    resample as resample,
    simulate_Matern as simulate_Matern,
    simulate_NeymanScott as simulate_NeymanScott,
    simulate_Poisson as simulate_Poisson,
    simulate_Thomas as simulate_Thomas,
    simulate_cluster as simulate_cluster,
    simulate_dstorm as simulate_dstorm,
    simulate_frame_numbers as simulate_frame_numbers,
    simulate_tracks as simulate_tracks,
    simulate_uniform as simulate_uniform,
)
from .tests import test as test
from .utils import (
    biased_variance as biased_variance,
    dependency_info as dependency_info,
    iterate_2d_array as iterate_2d_array,
    ratio_fwhm_to_sigma as ratio_fwhm_to_sigma,
    show_versions as show_versions,
    system_info as system_info,
    weighted_mean_variance as weighted_mean_variance,
)
from .visualize import (
    Colormap as Colormap,
    Colormaps as Colormaps,
    HistogramEqualization as HistogramEqualization,
    Trafo as Trafo,
    adjust_contrast as adjust_contrast,
    apply_window as apply_window,
    colormap_registry as colormap_registry,
    get_colormap as get_colormap,
    get_rois as get_rois,
    render_2d as render_2d,
    render_2d_mpl as render_2d_mpl,
    render_2d_napari as render_2d_napari,
    render_2d_napari_image as render_2d_napari_image,
    render_2d_rgb_mpl as render_2d_rgb_mpl,
    render_2d_rgb_napari as render_2d_rgb_napari,
    render_2d_scatter_density as render_2d_scatter_density,
    render_3d as render_3d,
    render_3d_napari as render_3d_napari,
    render_3d_napari_image as render_3d_napari_image,
    render_3d_rgb_napari as render_3d_rgb_napari,
    save_rois as save_rois,
    scatter_2d_mpl as scatter_2d_mpl,
    scatter_3d_mpl as scatter_3d_mpl,
    select_by_drawing_napari as select_by_drawing_napari,
)

__all__: list[str]
__version__: str
locdata_id: int



"""

File input/output for localization data in ASDF files

"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsWrite

import pandas as pd
from asdf import AsdfFile
from asdf import open as asdf_open
from google.protobuf import json_format

from locan.constants import PROPERTY_KEYS
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)

__all__: list[str] = ["save_asdf", "load_asdf_file"]

logger = logging.getLogger(__name__)


def save_asdf(
    locdata: LocData, path: str | os.PathLike[Any] | SupportsWrite[Any]
) -> None:
    """
    Save LocData attributes in an asdf file.

    In the Advanced Scientific Data Format (ASDF) file format we store
    metadata, properties and column names as human-readable yaml header.
    The data is stored as binary numpy.ndarray.

    Note
    ----
    Only selected LocData attributes are saved.
    Currently these are: 'data', 'columns', 'properties', 'meta'.

    Parameters
    ----------
    locdata
        The LocData object to be saved.
    path
        File path including file name to save to.
    """
    # Prepare tree
    meta_json = json_format.MessageToJson(
        locdata.meta, always_print_fields_with_no_presence=False
    )
    tree = {
        "data": locdata.data.values,
        "columns": list(locdata.data),
        "properties": locdata.properties,
        "meta": meta_json,
    }

    # Create the ASDF file object from tree
    af = AsdfFile(tree)

    # Write the data to a new file
    af.write_to(path)


def load_asdf_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
) -> LocData:
    """
    Load data from ASDF localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file.
        None means that all available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with asdf_open(path) as af:
        new_df = pd.DataFrame(
            {
                k: af.tree["data"][slice(nrows), n]
                for n, k in enumerate(af.tree["columns"])
            }
        )

        column_keys = convert_property_names(
            properties=new_df.columns.tolist(), property_mapping=None
        )
        mapper = {key: value for key, value in zip(new_df.columns, column_keys)}
        new_df = new_df.rename(columns=mapper)

        if convert:
            new_df = convert_property_types(new_df, types=PROPERTY_KEYS)

        locdata = LocData(dataframe=new_df)
        locdata.meta = json_format.Parse(
            af.tree["meta"], locdata.meta, ignore_unknown_fields=True
        )
        locdata.meta.file.type = metadata_pb2.ASDF
        locdata.meta.file.path = str(path)
    return locdata



"""

Utility functions.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   miscellaneous
   statistics
   system_information

"""

from __future__ import annotations

from locan.utils import miscellaneous, statistics, system_information

from .miscellaneous import *
from .statistics import *
from .system_information import *

__all__: list[str] = []
__all__.extend(miscellaneous.__all__)
__all__.extend(statistics.__all__)
__all__.extend(system_information.__all__)



from copy import deepcopy

import matplotlib.pyplot as plt  # needed for visual inspection  # noqa: F401
import pytest

from locan import (
    LocalizationProperty2d,
    LocData,
    Rectangle,
    render_2d_mpl,  # needed for visual inspection  # noqa: F401
    simulate_uniform,
)
from locan.analysis.localization_property_2d import _gauss_2d


def test_Localization_property_2d_empty(caplog):
    lprop = LocalizationProperty2d().compute(LocData())
    lprop.plot()
    lprop.plot_residuals()
    lprop.plot_deviation_from_mean()
    lprop.plot_deviation_from_median()
    lprop.report()
    assert caplog.record_tuples == [
        ("locan.analysis.localization_property_2d", 30, "Locdata is empty."),
        ("locan.analysis.localization_property_2d", 30, "No results available"),
    ]


@pytest.mark.visual
def test_Localization_property_2d_simulated_data(capfd):
    locdata = simulate_uniform(n_samples=10_000, region=Rectangle((0, 500), 1000, 500))
    intensity = _gauss_2d(
        x=locdata.data.position_x,
        y=locdata.data.position_y,
        amplitude=1000,
        center_x=800,
        center_y=700,
        sigma_x=300,
        sigma_y=300,
    )
    df = locdata.dataframe.assign(intensity=intensity)
    locdata = locdata.update(dataframe=df)

    # render_2d_mpl(locdata, other_property="intensity", bin_size=20)

    lprop = LocalizationProperty2d(
        meta=None, other_property="intensity", bin_size=(20, 50)
    ).compute(locdata)
    assert "model_result" in lprop.results._fields
    assert lprop.results.model_result.params

    lprop.report()
    captured = capfd.readouterr()
    assert captured.out[:16] == "Fit results for:"

    lprop.plot()
    lprop.plot_residuals()
    lprop.plot_deviation_from_mean()
    lprop.plot_deviation_from_median()
    plt.show()

    plt.close("all")


def test_Localization_property_2d(capfd, locdata_blobs_2d):
    locdata = deepcopy(locdata_blobs_2d)
    intensity = _gauss_2d(
        x=locdata.data.position_x,
        y=locdata.data.position_y,
        amplitude=1000,
        center_x=800,
        center_y=700,
        sigma_x=300,
        sigma_y=300,
    )
    df = locdata.dataframe.assign(intensity=intensity)
    locdata = locdata.update(dataframe=df)

    lprop = LocalizationProperty2d(
        meta=None, other_property="intensity", bin_size=20
    ).compute(locdata)
    assert "model_result" in lprop.results._fields
    assert lprop.results.model_result.params

    lprop.report()
    captured = capfd.readouterr()
    assert captured.out[:16] == "Fit results for:"

    lprop.plot(colors="r")
    lprop.plot_residuals()
    lprop.plot_deviation_from_mean()
    lprop.plot_deviation_from_median()
    # plt.show()

    plt.close("all")



"""
Benchmark functions for imports
"""


def timeraw_import_locan():
    """Benchmark import time for locan"""
    return """
    import locan
    """


def timeraw_import_locan_analysis():
    """Benchmark import time for locan"""
    return """
    import locan.analysis
    """


def timeraw_import_locan_configuration():
    """Benchmark import time for locan"""
    return """
    import locan.configuration
    """


def timeraw_import_locan_constants():
    """Benchmark import time for locan"""
    return """
    import locan.constants
    """


def timeraw_import_locan_data():
    """Benchmark import time for locan"""
    return """
    import locan.data
    """


def timeraw_import_locan_datasets():
    """Benchmark import time for locan"""
    return """
    import locan.datasets
    """


def timeraw_import_locan_dependencies():
    """Benchmark import time for locan"""
    return """
    import locan.dependencies
    """


def timeraw_import_locan_gui():
    """Benchmark import time for locan"""
    return """
    import locan.gui
    """


def timeraw_import_locan_locan_io():
    """Benchmark import time for locan"""
    return """
    import locan.locan_io
    """


def timeraw_import_locan_rois():
    """Benchmark import time for locan"""
    return """
    import locan.rois
    """


def timeraw_import_locan_scripts():
    """Benchmark import time for locan"""
    return """
    import locan.scripts
    """


def timeraw_import_locan_simulation():
    """Benchmark import time for locan"""
    return """
    import locan.simulation
    """


def timeraw_import_locan_tests():
    """Benchmark import time for locan"""
    return """
    import locan.tests
    """


def timeraw_import_locan_utils():
    """Benchmark import time for locan"""
    return """
    import locan.utils
    """


def timeraw_import_locan_visualize():
    """Benchmark import time for locan"""
    return """
    import locan.visualize
    """


def timeraw_import_asdf():
    """Benchmark import time"""
    return """
    import asdf
    """


def timeraw_import_boost_histogram():
    """Benchmark import time"""
    return """
    import boost_histogram
    """


def timeraw_import_fast_histogram():
    """Benchmark import time"""
    return """
    import fast_histogram
    """


def timeraw_import_lmfit():
    """Benchmark import time"""
    return """
    import lmfit
    """


def timeraw_import_matplotlib_pyplot():
    """Benchmark import time"""
    return """
    import matplotlib.pyplot
    """


def timeraw_import_networkx():
    """Benchmark import time"""
    return """
    import networkx
    """


def timeraw_import_numpy():
    """Benchmark import time for numpy"""
    return """
    import numpy
    """


def timeraw_import_pandas():
    """Benchmark import time for pandas"""
    return """
    import pandas
    """


def timeraw_import_protobuf():
    """Benchmark import time"""
    return """
    import google.protobuf
    """


def timeraw_import_qtpy():
    """Benchmark import time"""
    return """
    import qtpy
    """


def timeraw_import_ruamel_yaml():
    """Benchmark import time"""
    return """
    import ruamel.yaml
    """


def timeraw_import_skimage():
    """Benchmark import time"""
    return """
    import skimage
    """


def timeraw_import_sklearn():
    """Benchmark import time"""
    return """
    import sklearn
    """


def timeraw_import_scipy():
    """Benchmark import time"""
    return """
    import scipy
    """


def timeraw_import_shapely():
    """Benchmark import time"""
    return """
    import shapely
    """


def timeraw_import_tifffile():
    """Benchmark import time"""
    return """
    import tifffile
    """


def timeraw_import_tqdm():
    """Benchmark import time"""
    return """
    import tqdm
    """



from io import StringIO

import locan.constants
from locan import FileType
from locan.locan_io import load_Elyra_file
from locan.locan_io.locdata.elyra_io import load_Elyra_header


def test_get_correct_column_names_from_Elyra_header():
    columns = load_Elyra_header(
        path=locan.ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    )
    assert columns == [
        "original_index",
        "frame",
        "frames_number",
        "frames_missing",
        "position_x",
        "position_y",
        "uncertainty",
        "intensity",
        "local_background_sigma",
        "chi_square",
        "psf_half_width",
        "channel",
        "slice_z",
    ]

    file_like = StringIO(
        "Index	First Frame	Number Frames	Frames Missing	Position X [nm]	Position Y [nm]\n"
        "1  1   1   0   15850.6 23502.1"
    )
    columns = load_Elyra_header(path=file_like)
    assert columns == [
        "original_index",
        "frame",
        "frames_number",
        "frames_missing",
        "position_x",
        "position_y",
    ]


def test_loading_Elyra_file():
    file_path = locan.ROOT_DIR / "tests/test_data/Elyra_dstorm_data.txt"
    dat = load_Elyra_file(path=file_path)
    # loading is not limited by nrows=10 to ensure correct treatment of file appendix and NUL character.
    assert len(dat) == 15
    assert dat.meta.file.type == FileType.ELYRA.value
    assert dat.meta.file.path == str(file_path)

    file_like = StringIO(
        "Index\tFirst Frame\tNumber Frames\tFrames Missing\tPosition X [nm]\tPosition Y [nm]\n"
        "1\t1\t1\t0\t15850.6\t23502.1"
    )
    dat = load_Elyra_file(path=file_like)
    # loading is not limited by nrows=10 to ensure correct treatment of file appendix and NUL character.
    assert len(dat) == 1



from io import StringIO

import numpy as np

import locan.constants
from locan.locan_io import load_rapidSTORM_file, load_rapidSTORM_track_file
from locan.locan_io.locdata.rapidstorm_io import (
    load_rapidSTORM_header,
    load_rapidSTORM_track_header,
)


def test_get_correct_column_names_from_rapidSTORM_header():
    columns = load_rapidSTORM_header(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    )
    assert columns == [
        "position_x",
        "position_y",
        "frame",
        "intensity",
        "chi_square",
        "local_background",
    ]

    file_like = StringIO(
        '# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations>\n'
        "9657.4 24533.5 0 33290.1 1.19225e+006 767.733"
    )
    columns = load_rapidSTORM_header(path=file_like)
    assert columns == [
        "position_x",
        "position_y",
        "frame",
        "intensity",
        "chi_square",
        "local_background",
    ]


def test_loading_rapidSTORM_file():
    path = locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_data.txt"
    dat = load_rapidSTORM_file(path=path, nrows=10)
    # print(dat.data.head())
    # dat.print_meta()
    assert len(dat) == 10
    assert dat.meta.file.type == locan.FileType.RAPIDSTORM.value
    assert dat.meta.file.path == str(path)

    file_like = StringIO(
        '# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="3.27165e-005 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations>\n'
        "9657.4 24533.5 0 33290.1 1.19225e+006 767.733"
    )
    dat = load_rapidSTORM_file(path=file_like, nrows=1)
    assert len(dat) == 1


def test_get_correct_column_names_from_rapidSTORM_track_header():
    columns = load_rapidSTORM_track_header(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_track_data.txt"
    )
    assert columns == (
        ["position_x", "position_y", "frame", "intensity"],
        [
            "position_x",
            "uncertainty_x",
            "position_y",
            "uncertainty_y",
            "frame",
            "intensity",
            "chi_square",
            "local_background",
        ],
    )

    file_like = StringIO(
        '# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-0-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in X" unit="nanometer" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in Y" unit="nanometer" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations></localizations>\n'
        "5417.67 8439.85 1 4339.29 2 5421.22 25 8440.4 25 0 2292.39 48696.1 149.967 5414.13 25 8439.3 25 1 2046.9 65491.4 142.521"
    )
    columns = load_rapidSTORM_track_header(path=file_like)
    assert columns == (
        ["position_x", "position_y", "frame", "intensity"],
        [
            "position_x",
            "uncertainty_x",
            "position_y",
            "uncertainty_y",
            "frame",
            "intensity",
            "chi_square",
            "local_background",
        ],
    )


def test_loading_rapidSTORM_track_file():
    dat = load_rapidSTORM_track_file(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_track_data.txt",
        min_localization_count=2,
        nrows=10,
    )
    # print(dat.data.head())
    # print(dat.data.columns)
    # dat.print_meta()
    assert np.array_equal(
        dat.data.columns,
        [
            "localization_count",
            "position_x",
            "uncertainty_x",
            "position_y",
            "uncertainty_y",
            "intensity",
            "local_background",
            "frame",
            "region_measure_bb",
            "localization_density_bb",
            "subregion_measure_bb",
        ],
    )
    assert len(dat) == 9
    # len(dat) is 9 and not 10 since one row is filtered out y min_localization_count=2

    dat = load_rapidSTORM_track_file(
        path=locan.ROOT_DIR / "tests/test_data/rapidSTORM_dstorm_track_data.txt",
        collection=False,
        nrows=10,
    )
    assert np.array_equal(
        dat.data.columns, ["position_x", "position_y", "frame", "intensity"]
    )
    assert len(dat) == 10

    file_like = StringIO(
        '# <localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><localizations insequence="true" repetitions="variable"><field identifier="Position-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in X" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-0-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in X" unit="nanometer" /><field identifier="Position-1-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position in sample space in Y" unit="nanometer" min="0 m" max="8.442e-006 m" /><field identifier="Position-1-0-uncertainty" syntax="floating point with . for decimals and optional scientific e-notation" semantic="position uncertainty in sample space in Y" unit="nanometer" /><field identifier="ImageNumber-0-0" syntax="integer" semantic="frame number" unit="frame" min="0 fr" /><field identifier="Amplitude-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="emission strength" unit="A/D count" /><field identifier="FitResidues-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="fit residue chi square value" unit="dimensionless" /><field identifier="LocalBackground-0-0" syntax="floating point with . for decimals and optional scientific e-notation" semantic="local background" unit="A/D count" /></localizations></localizations>\n'
        "5417.67 8439.85 1 4339.29 2 5421.22 25 8440.4 25 0 2292.39 48696.1 149.967 5414.13 25 8439.3 25 1 2046.9 65491.4 142.521"
    )
    dat = load_rapidSTORM_file(path=file_like, nrows=1)
    assert len(dat) == 1



import numpy as np

from locan import LocData
from locan.locan_types import LocData as LocDataProtocol
from locan.locan_types import RandomGeneratorSeed


def test_types() -> None:
    seed: RandomGeneratorSeed = np.array([1, 2])  # type: ignore
    # raises mypy assignment error
    assert isinstance(seed, np.ndarray)

    seed = [1, 2]
    assert isinstance(seed, list)


def test_locdata() -> None:
    locdata: LocDataProtocol = LocData()
    # LocDataProtocol is not runtime checkable:
    # assert isinstance(locdata, LocDataProtocol)
    assert isinstance(locdata, LocData)



"""

Analyze geometrical properties of the convex hull of localization coordinates.

Localization coordinates in localization clusters come with a certain variance.
In resolution-limited clusters the variance is determined by the
localization precision. Accordingly, the geometrical properties of convex hulls
also vary and can be analyzed as function of
localization counts to help characterize localization clusters [1]_.

References
----------
.. [1] Ebert V, Eiring P, Helmerich DA, Seifert R, Sauer M, Doose S.
   Convex hull as diagnostic tool in single-molecule localization microscopy.
   Bioinformatics 38(24), 2022, 5421-5429, doi: 10.1093/bioinformatics/btac700.

"""

# todo add fit procedure to estimate variance_estimate
from __future__ import annotations

import importlib.resources as importlib_resources
import logging
import sys
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Literal, NamedTuple

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from locan.data.locdata import LocData

import boost_histogram as bh
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.integrate as integrate
import scipy.special as special

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.process.aggregate import Bins

if TYPE_CHECKING:
    import matplotlib as mpl

__all__: list[str] = ["ConvexHullExpectation", "ConvexHullExpectationBatch"]

logger = logging.getLogger(__name__)


class ConvexHullProperty(Enum):
    REGION_MEASURE_2D = auto()
    SUBREGION_MEASURE_2D = auto()
    REGION_MEASURE_3D = auto()
    SUBREGION_MEASURE_3D = auto()
    # might be extended -
    # but the previous entries must correspond to ConvexHullExpectationResource


ConvexHullExpectationResource = dict(
    REGION_MEASURE_2D="lookup_table_area_2d.npy",
    SUBREGION_MEASURE_2D="lookup_table_peri_2d.npy",
    REGION_MEASURE_3D="lookup_table_vol_3d.npy",
    SUBREGION_MEASURE_3D="lookup_table_area_3d.npy",
)


class ConvexHullExpectationValues(NamedTuple):
    n_points: list[int] | npt.NDArray[np.int64]
    expectation: npt.NDArray[np.int64 | np.float64]
    std_pos: npt.NDArray[np.float64]
    std_neg: npt.NDArray[np.float64]


def _get_resource(
    resource_directory: str, resource: str
) -> ConvexHullExpectationValues:
    """
    Get convex hull property values from resource files produced by
    numerical simulations.

    Note
    -----
    Original data can be found at
    https://github.com/super-resolution/Ebert-et-al-2022-supplement

    Parameters
    ----------
    resource_directory
        Directory with resources
    resource
        Name of resource file within resource_directory

    Returns
    -------
    ConvexHullExpectationValues
    """

    try:
        resource_ = importlib_resources.files(resource_directory).joinpath(resource)
        resource_values = np.load(str(resource_))

    except (AttributeError, TypeError):  # required for python < 3.9
        with importlib_resources.path(
            package=resource_directory, resource=resource
        ) as resource_:
            resource_values = np.load(str(resource_))

    if "2d" in resource:  # hard coded corresponding to n_points in resources
        n_points = list(range(3, 201))
    elif "3d" in resource:
        n_points = list(range(4, 201))
    else:
        raise ValueError("resource is neither '2d' nor '3d'")
    if len(n_points) != resource_values.shape[1]:
        raise ValueError("The resource files are not correct.")
    return ConvexHullExpectationValues(
        n_points=n_points,
        expectation=resource_values[0],
        std_pos=resource_values[1],
        std_neg=resource_values[2],
    )


# The algorithms


def compute_convex_hull_region_measure_2d(n_points: int, sigma: float = 1) -> float:
    """
    Compute the expected convex hull area for `n_points` data points
    that are normal distributed in 2-dimensional space
    with standard deviation `sigma`.

    Parameters
    ----------
    n_points
        Number of points of convex hull
    sigma
        Standard deviation of normal-distributed point coordinates

    Returns
    -------
    float
        Expectation value for convex hull area
    """

    def pdf(x: float) -> float:
        return_value = 1 / (np.sqrt(2 * np.pi)) * np.exp((-1 / 2) * (x**2))
        return float(return_value)

    def cdf(x: float) -> float:
        return_value = 1 / 2 * (1 + special.erf(x / np.sqrt(2)))
        return float(return_value)

    try:
        area = (
            3
            * np.pi
            * special.binom(n_points, 3)
            * integrate.quad(
                lambda x: cdf(x) ** (n_points - 3) * pdf(x) ** 3, -np.inf, np.inf
            )[0]
        )
        return_value = sigma**2 * area
    except ZeroDivisionError:
        return_value = np.nan

    return float(return_value)


def _get_convex_hull_property_expectation(
    convex_hull_property: str | ConvexHullProperty,
    n_points: npt.ArrayLike,
    sigma: float = 1,
) -> ConvexHullExpectationValues:
    """
    Get the expected convex hull property for `n_points` data points
    that are normal distributed in 2- or 3-dimensional space
    with standard deviation `sigma`.

    Parameters
    ----------
    convex_hull_property
        Choose property and dimension of convex hull
    n_points
        Number of points of convex hull
    sigma
        Standard deviation of normal-distributed point coordinates

    Returns
    -------
    ConvexHullExpectationValues
        Expectation values for convex hull property
    """
    try:
        convex_hull_property = convex_hull_property.name.upper()  # type: ignore
    except AttributeError:
        convex_hull_property = convex_hull_property.upper()  # type: ignore
    convex_hull_expectation_values = _get_resource(
        resource_directory="locan.analysis.resources.convex_hull_expectation",
        resource=ConvexHullExpectationResource[convex_hull_property],
    )
    n_points, _, indices = np.intersect1d(
        n_points, convex_hull_expectation_values.n_points, return_indices=True
    )

    if convex_hull_property == "REGION_MEASURE_2D":
        factor = sigma**2
    elif convex_hull_property == "SUBREGION_MEASURE_2D":
        factor = sigma
    elif convex_hull_property == "REGION_MEASURE_3D":
        factor = sigma**3
        logger.warning("The expectation is scaled by sigma^2")
    elif convex_hull_property == "SUBREGION_MEASURE_3D":
        factor = sigma**2
        logger.warning("The expectation is scaled by sigma^2")
    else:
        raise ValueError("convex_hull_property is undefined")

    result = ConvexHullExpectationValues(
        n_points=n_points,
        expectation=convex_hull_expectation_values.expectation[indices] * factor,  # type: ignore[arg-type]
        std_pos=convex_hull_expectation_values.std_pos[indices] * factor,
        std_neg=convex_hull_expectation_values.std_neg[indices] * factor,
    )
    return result


# The specific analysis classes


@dataclass(repr=False)
class ConvexHullExpectationResults:
    values: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being reference_index
    # with columns: loc_property, other_loc_property, value_to_expectation_ratio
    grouped: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being other_loc_property
    # with columns: loc_property_mean, loc_property_std, expectation


class ConvexHullExpectation(_Analysis):
    """
    Analyze geometrical properties of the convex hull of localization
    coordinates in relation to expected values.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    convex_hull_property : Literal["region_measure_ch", "subregion_measure_ch"]
        One of 'region_measure_ch' (i.e. area or volume)
        or 'subregion_measure_ch' (i.e. circumference or surface.)
    expected_variance : float | Iterable[float] | None
        The expected variance for all or each localization property.
        The expected variance equals the squared localization precision
        for localization position coordinates.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict[str, Any]
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : ConvexHullExpectationResults | None
        Computed results.
    distribution_statistics : dict[str, Any] | None
        Distribution parameters derived from MLE fitting of results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        convex_hull_property: Literal[
            "region_measure_ch", "subregion_measure_ch"
        ] = "region_measure_ch",
        expected_variance: float | Iterable[float] | None = None,
    ) -> None:
        if convex_hull_property not in ["region_measure_ch", "subregion_measure_ch"]:
            raise TypeError(
                "convex_hull_property must be one of "
                "[region_measure_ch, subregion_measure_ch]"
            )
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.expected_variance = expected_variance
        self.results: ConvexHullExpectationResults | None = None
        self.distribution_statistics: dict[str, Any] | None = None
        self._dimension: int | None = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = ConvexHullExpectationResults()
        assert self.results is not None  # type narrowing # noqa: S101
        loc_property = self.parameter["convex_hull_property"]
        self._dimension = locdata.dimension

        try:
            self.results.values = locdata.data.loc[
                :, ["localization_count", loc_property]  # type: ignore[list-item]
            ]
        except KeyError:
            locdata.update_convex_hulls_in_references()
            self.results.values = locdata.data.loc[
                :, ["localization_count", loc_property]  # type: ignore[list-item]
            ]

        grouped = self.results.values.groupby("localization_count")

        self.results.grouped[loc_property + "_mean"] = grouped.mean()
        self.results.grouped[loc_property + "_std"] = grouped.std()

        convex_hull_property_ = loc_property[:-2] + f"{self._dimension}d"
        convex_hull_property_ = ConvexHullProperty[convex_hull_property_.upper()]

        if self.expected_variance is None:
            self.results.grouped["expectation"] = pd.NA
            self.results.grouped["expectation_std_pos"] = pd.NA
            self.results.grouped["expectation_std_neg"] = pd.NA
        else:
            convex_hull_expectation_values = _get_convex_hull_property_expectation(
                n_points=self.results.grouped.index,
                convex_hull_property=convex_hull_property_,
                sigma=np.sqrt(self.expected_variance),  # type: ignore[arg-type]
            )
            new_dict = dict(
                expectation=convex_hull_expectation_values.expectation,
                expectation_std_pos=convex_hull_expectation_values.std_pos,
                expectation_std_neg=convex_hull_expectation_values.std_neg,
            )
            new_df = pd.DataFrame(
                new_dict, index=convex_hull_expectation_values.n_points
            )
            self.results.grouped = pd.merge(
                self.results.grouped,
                new_df,
                how="outer",
                left_index=True,
                right_index=True,
            )

        self.results.values = pd.merge(
            self.results.values,
            self.results.grouped["expectation"],
            how="outer",
            left_on="localization_count",
            right_index=True,
        )
        self.results.values["value_to_expectation_ratio"] = (
            self.results.values[loc_property] / self.results.values["expectation"]
        )

        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        convex_hull_property as function of localization counts.

        Parameters
        ----------
        ax
            The axes on which to show the image
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        self.results.values.plot(
            kind="scatter",
            alpha=0.2,
            x="localization_count",
            y=self.parameter["convex_hull_property"],
            color="gray",
            ax=ax,
            **dict(dict(legend=False), **kwargs),
        )

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["convex_hull_property"] + "_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="black",
                ax=ax,
            )

        if not self.results.grouped.expectation_std_pos.isna().all():
            df = (
                self.results.grouped.expectation
                + self.results.grouped.expectation_std_pos
            )
            df.plot(
                kind="line",
                linestyle=":",
                color="gray",
                ax=ax,
            )

        if not self.results.grouped.expectation_std_neg.isna().all():
            df = (
                self.results.grouped.expectation
                - self.results.grouped.expectation_std_neg
            )
            df.plot(
                kind="line",
                linestyle=":",
                color="gray",
                ax=ax,
            )

        ax.set(
            title="Convex Hull Expectation",
            xlabel="localization_count",
            ylabel=self.parameter["convex_hull_property"],
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | None
        ) = None,
        log: bool = True,
        fit: bool = False,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        2-dimensional histogram of convex_hull_property
        and localization counts.

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            The bin specification as defined in :class:`Bins`
        bin_edges
            Bin edges for all or each dimension
            with shape (dimension, n_bin_edges).
        bin_range
            Minimum and maximum edge for all or each dimensions
            with shape (2,) or (dimension, 2).
        n_bins
            The number of bins for all or each dimension.
            5 yields 5 bins in all dimensions.
            (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
        bin_size
            The size of bins for all or each bin and for all or each dimension
            with shape (dimension,) or (dimension, n_bins).
            5 would describe bin_size of 5 for all bins in all dimensions.
            ((2, 5),) yield bins of size (2, 5) for one dimension.
            (2, 5) yields bins of size 2 for one dimension and 5 for the other
            dimension.
            ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
            (1, 3) for the other dimension.
            To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
        log
            Flag for plotting on a log scale.
        fit
            Flag indicating if distribution fit is shown.
            The fit will only be computed if `distribution_statistics`
            is None.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.pcolormesh`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        fig = ax.get_figure()

        other_loc_property = "localization_count"
        loc_property = self.parameter["convex_hull_property"]

        if all(
            item_ is None for item_ in [bins, n_bins, bin_size, bin_edges, bin_range]
        ):
            max_other_loc_property_ = self.results.grouped.index.max()
            max_loc_property_ = self.results.values[loc_property].max()

            if log:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_,
                            3,
                            max_other_loc_property_,
                            transform=bh.axis.transform.log,
                        ),
                        bh.axis.Regular(
                            50, 1, max_loc_property_, transform=bh.axis.transform.log
                        ),
                    )
                )
            else:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_, 3, max_other_loc_property_
                        ),
                        bh.axis.Regular(50, 1, max_loc_property_),
                    )
                )
            bins = Bins(bins=axes)
        else:
            try:
                bins = Bins(
                    bins,
                    n_bins,
                    bin_size,
                    bin_edges,
                    bin_range,
                    labels=[loc_property, other_loc_property],
                )
            except ValueError as exc:
                # todo: check if message is appropriate
                # the error is raised again only to adapt the message.
                raise ValueError(
                    "Bin dimension and len of `loc_properties` is incompatible."
                ) from exc
        axes = bins.boost_histogram_axes
        histogram = bh.Histogram(*axes)
        histogram.reset()
        histogram.fill(
            self.results.values[other_loc_property], self.results.values[loc_property]
        )
        mesh = ax.pcolormesh(*histogram.axes.edges.T, histogram.view().T, **kwargs)
        fig.colorbar(mesh)  # type:ignore[union-attr]

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["convex_hull_property"] + "_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="white",
                ax=ax,
            )

        if not self.results.grouped.expectation_std_pos.isna().all():
            df = (
                self.results.grouped.expectation
                + self.results.grouped.expectation_std_pos
            )
            df.plot(
                kind="line",
                linestyle=":",
                color="gray",
                ax=ax,
            )

        if not self.results.grouped.expectation_std_neg.isna().all():
            df = (
                self.results.grouped.expectation
                - self.results.grouped.expectation_std_neg
            )
            df.plot(
                kind="line",
                linestyle=":",
                color="gray",
                ax=ax,
            )

        ax.set(
            title="Convex Hull Expectation",
            xlabel=other_loc_property,
            ylabel=loc_property,
        )
        if log:
            ax.set(xscale="log", yscale="log")

        if fit:
            raise NotImplementedError

        return ax


class ConvexHullExpectationBatch(_Analysis):
    """
    Analyze geometrical properties of the convex hull of localization
    coordinates in relation to expected values.

    See Also
    --------
    :class:`ConvexHullExpectation`

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    convex_hull_property : Literal['region_measure_ch', 'subregion_measure_ch']
        One of 'region_measure_ch' (i.e. area or volume)
        or 'subregion_measure_ch' (i.e. circumference or surface.)
    expected_variance : float | Iterable[float] | None
        The expected variance for all or each localization property.
        The expected variance equals the squared localization precision
        for localization position coordinates.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : ConvexHullExpectationResults
        Computed results.
    distribution_statistics : Distribution_stats, None
        Distribution parameters derived from MLE fitting of results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        convex_hull_property: Literal[
            "region_measure_ch", "subregion_measure_ch"
        ] = "region_measure_ch",
        expected_variance: float | Iterable[float] | None = None,
    ) -> None:
        if convex_hull_property not in ["region_measure_ch", "subregion_measure_ch"]:
            raise TypeError(
                "convex_hull_property must be one of "
                "[region_measure_ch, subregion_measure_ch]"
            )
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.expected_variance = expected_variance
        self.results = None
        self.batch = None
        self._dimension: int | None = None
        self._class = ConvexHullExpectation(
            convex_hull_property=convex_hull_property,
            expected_variance=expected_variance,
        )
        self.distribution_statistics = None

    def compute(self, locdatas: Iterable[LocData]) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdatas : Iterable[LocData]
            Localization data.

        Returns
        -------
        Self
        """
        convex_hull_expectation_batch: list[ConvexHullExpectation] = []
        dimensions: list[int | None] = []
        for locdata_ in locdatas:
            che = ConvexHullExpectation(
                convex_hull_property=self.parameter["convex_hull_property"],
                expected_variance=self.parameter["expected_variance"],
            ).compute(locdata_)
            convex_hull_expectation_batch.append(che)
            dimensions.append(che._dimension)

        if bool(locdatas):
            dimensions_ = set(dimensions)
            if len(dimensions_) == 1:  # check if all are equal
                self._dimension = dimensions_.pop()
            else:
                raise ValueError("The dimensions of all locdata must be the same.")

        self.from_batch(batch=convex_hull_expectation_batch)
        return self

    def from_batch(
        self, batch: Iterable[ConvexHullExpectation], dimension: int | None = None
    ) -> Self:
        if not bool(batch) or all(item_.results is None for item_ in batch):
            logger.warning("The batch is empty.")
            return self

        loc_property = self.parameter["convex_hull_property"]
        if self._dimension is None:
            if dimension is not None:
                self._dimension = dimension
            else:
                dimension_ = set(item_._dimension for item_ in batch)
                if len(dimension_) == 1:
                    self._dimension = dimension_.pop()

        self.results = ConvexHullExpectationResults()
        self.results.values = pd.concat(
            [item_.results.values for item_ in batch if item_.results is not None],
            ignore_index=True,
        )

        grouped = self.results.values.groupby("localization_count")

        self.results.grouped.loc[:, loc_property + "_mean"] = grouped[
            loc_property
        ].mean()
        self.results.grouped.loc[:, loc_property + "_std"] = grouped[loc_property].std()

        convex_hull_property_ = loc_property[:-2] + f"{self._dimension}d"
        convex_hull_property_ = ConvexHullProperty[convex_hull_property_.upper()]

        if self.expected_variance is None:
            self.results.grouped.loc[:, "expectation"] = pd.NA  # type: ignore
            self.results.grouped.loc[:, "expectation_std_pos"] = pd.NA  # type: ignore
            self.results.grouped.loc[:, "expectation_std_neg"] = pd.NA  # type: ignore
        else:
            convex_hull_expectation_values = _get_convex_hull_property_expectation(
                n_points=self.results.grouped.index,
                convex_hull_property=convex_hull_property_,
                sigma=np.sqrt(self.expected_variance),  # type: ignore
            )
            new_dict = dict(
                expectation=convex_hull_expectation_values.expectation,
                expectation_std_pos=convex_hull_expectation_values.std_pos,
                expectation_std_neg=convex_hull_expectation_values.std_neg,
            )
            new_df = pd.DataFrame(
                new_dict, index=convex_hull_expectation_values.n_points
            )
            self.results.grouped = pd.merge(
                self.results.grouped,
                new_df,
                how="outer",
                left_index=True,
                right_index=True,
            )
        self._class.results = self.results
        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        return self._class.plot(ax=ax, **kwargs)

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | None
        ) = None,
        log: bool = True,
        fit: bool = False,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        return self._class.hist(
            ax=ax,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range,
            log=log,
            fit=fit,
            **kwargs,
        )



from locan.scripts.utilities import _type_converter_rescale


def test__type_converter_rescale():
    return_value = _type_converter_rescale(input_string="None")
    assert return_value is None
    return_value = _type_converter_rescale(input_string="True")
    assert return_value is True
    return_value = _type_converter_rescale(input_string="False")
    assert return_value is False
    return_value = _type_converter_rescale(input_string="EQUALIZE")
    assert return_value == "EQUALIZE"
    return_value = _type_converter_rescale(input_string="0.2 0.9")
    assert return_value == (0.2, 0.9)
    return_value = _type_converter_rescale(input_string="0.2, 0.9")
    assert return_value == (0.2, 0.9)
    return_value = _type_converter_rescale(input_string="0.2 0.9 1")
    assert return_value == (0.2, 0.9)



"""

File input/output for localization data in SMAP files.

"""

from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsWrite

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)

if HAS_DEPENDENCY["h5py"]:
    import h5py


__all__: list[str] = ["load_SMAP_header", "load_SMAP_file", "save_SMAP_csv"]

logger = logging.getLogger(__name__)


@needs_package("h5py")
def _read_SMAP_header(file: Mapping[str, Any]) -> list[str]:
    """
    Identify column names from a SMAP single-molecule localization file.

    Parameters
    ----------
    file
        HDF5 file object.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """
    # list identifiers
    identifiers = list(file["saveloc"]["loc"].keys())
    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.SMAP_KEYS
    )
    return column_keys


@needs_package("h5py")
def load_SMAP_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> list[str]:
    """
    Identify column names from a SMAP single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """
    with h5py.File(path, "r") as file:
        return _read_SMAP_header(file)


@needs_package("h5py")
def load_SMAP_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
) -> LocData:
    """
    Load data from a SMAP single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with h5py.File(path, "r") as file:
        columns = _read_SMAP_header(file)

        if file["saveloc"]["loc"]["frame"].shape == (0,):  # empty file
            logger.warning("File does not contain any data.")
            locdata = LocData()

        else:  # file not empty
            data = {}
            for key, value in file["saveloc"]["loc"].items():
                if value.shape is not None:
                    if key in locan.constants.SMAP_KEYS:
                        data[locan.constants.SMAP_KEYS[key]] = list(value[0][:nrows])
                    else:
                        data[key] = list(value[0][:nrows])

            dataframe = pd.DataFrame(data)

            if convert:
                dataframe = convert_property_types(
                    dataframe, types=locan.constants.PROPERTY_KEYS
                )

            locdata = LocData.from_dataframe(dataframe=dataframe)

    locdata.meta.source = metadata_pb2.EXPERIMENT
    locdata.meta.state = metadata_pb2.RAW
    locdata.meta.file.type = metadata_pb2.SMAP
    locdata.meta.file.path = str(path)

    for property_ in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        locdata.meta.localization_properties.add(
            name=property_, unit="nm", type="float"
        )

    del locdata.meta.history[:]
    locdata.meta.history.add(
        name="load_SMAP_file", parameter=f"path={path}, nrows={nrows}"
    )

    return locdata


def save_SMAP_csv(
    locdata: LocData,
    path: str | os.PathLike[Any] | SupportsWrite[Any],
) -> None:
    """
    Save LocData to SMAP-readable csv-file.

    In the csv-file file format we store only localization data with
    SMAP-readable column names.

    Parameters
    ----------
    locdata
        The LocData object to be saved.
    path
        File path including file name to save to.
    """
    # get data from locdata object
    dataframe = locdata.data

    # create reverse mapping to columns
    inv_map = {v: k for k, v in locan.constants.SMAP_KEYS.items()}

    # rename columns
    dataframe = dataframe.rename(index=str, columns=inv_map, inplace=False)
    valid_smap_columns = [
        key for key in dataframe.columns if key in locan.constants.SMAP_KEYS
    ]

    # write to csv
    dataframe[valid_smap_columns].to_csv(path, float_format="%.10g", index=False)  # type: ignore[arg-type]



"""
Benchmark functions for :mod:`locan.analysis.nearest_neighbor`
"""

import locan as lc
from locan.analysis.nearest_neighbor import (
    _nearest_neighbor_distances,
)


class BenchmarkNearestNeighborDistances:
    """
    Benchmarks for nearest neighbor computations
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)
        print(self.locdata.data.info())

    def check__nearest_neighbor_distances(self):
        results_df = _nearest_neighbor_distances(
            points=self.locdata.coordinates,
        )
        print(results_df)

    def time__nearest_neighbor_distances(self):
        _nearest_neighbor_distances(
            points=self.locdata.coordinates,
        )


def main():
    bm = BenchmarkNearestNeighborDistances()
    bm.setup()
    bm.check__nearest_neighbor_distances()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkNearestNeighborDistances()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(1000):
        bm.time__nearest_neighbor_distances()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("tottime")
    stats.print_stats("time_")


if __name__ == "__main__":
    main()
    main_profile()



"""

Utility functions for file input/output.

"""

from __future__ import annotations

import logging
import os
import re
from pathlib import Path

__all__: list[str] = ["find_file_upstream"]

logger = logging.getLogger(__name__)


def find_file_upstream(
    sub_directory: str | os.PathLike[str],
    pattern: str | None,
    regex: str | None = None,
    top_directory: str | os.PathLike[str] | None = None,
) -> Path | None:
    """
    Search for first upstream parent of sub_directory that contains pattern.
    Return first pattern found.
    Return None if no pattern has been found when parent equals directory.

    Parameters
    ----------
    sub_directory :
        Directory or file path to start with.
    pattern :
        glob pattern passed to :func:`Path.glob`
    regex :
        regex pattern passed to :func:`re.search` and applied in addition
        to glob pattern
    top_directory :
        Directory in which to stop the search.

    Returns
    -------
    Path | None
    """
    sub_directory = Path(sub_directory).resolve(strict=True)

    if pattern is None:
        pattern = "*.*"

    if top_directory is None:
        top_directory = sub_directory.anchor

    if regex is not None:
        regex_ = re.compile(regex)
    else:
        regex_ = None

    for parent in sub_directory.parents:
        file_list = list(parent.glob(pattern))

        if regex_ is not None:
            file_list = [
                file_ for file_ in file_list if regex_.search(str(file_)) is not None
            ]
        if file_list:
            return file_list[0]
        if parent == top_directory:
            return None
    return None



"""

Analyze localization property.

Localizations come with a range of properties including position coordinates,
emission strength, local background etc..
Most properties represent random variables that were drawn from an unknown
probability distribution.
It is often useful to analyze the properties from all localizations within a
selection and estimate the corresponding probability distribution.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import matplotlib.pyplot as plt
import numpy as np
from scipy import stats

if TYPE_CHECKING:
    import matplotlib as mpl
    import pandas as pd

    from locan.data.locdata import LocData

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis, _list_parameters

__all__: list[str] = ["LocalizationProperty"]

logger = logging.getLogger(__name__)


# The algorithms


def _localization_property(
    locdata: LocData, loc_property: str = "intensity", index: str | None = None
) -> pd.DataFrame:
    if index is None:
        results = locdata.data[[loc_property]]
    else:
        results = locdata.data[[loc_property, index]].set_index(index)
        results = results.sort_index()

    return results


# The specific analysis classes


class LocalizationProperty(_Analysis):
    """
    Analyze localization property with respect to probability density or
    variation over a specified index.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_property : str
        The property to analyze.
    index : str | None
        The property name that should serve as index (i.e. x-axis in x-y-plot)

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict[str, Any]
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Computed results.
    distribution_statistics : Distribution_stats | None
        Distribution parameters derived from MLE fitting of results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        loc_property: str = "intensity",
        index: str | None = None,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)

        self.results: pd.DataFrame | None = None
        self.distribution_statistics: _DistributionFits | None = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _localization_property(locdata=locdata, **self.parameter)
        return self

    def fit_distributions(
        self,
        distribution: str | stats.rv_continuous = stats.expon,
        with_constraints: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Fit probability density functions to the distributions of
        `loc_property` values in the results
        using MLE (scipy.stats).

        If with_constraints is true we put the following constraints on the
        fit procedure:
        If distribution is expon then
        `floc=np.min(self.analysis_class.results[self.loc_property].values)`.

        Parameters
        ----------
        distribution
            Distribution model to fit.
        with_constraints
            Flag to use predefined constraints on fit parameters.
        kwargs
            Other parameters are passed to `scipy.stat.distribution.fit()`.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(
                distribution, with_constraints=with_constraints, **kwargs
            )
        else:
            logger.warning("No results available to fit.")

    def plot(
        self, ax: mpl.axes.Axes | None = None, window: int = 1, **kwargs: Any
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        running average of results over window size.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to show the image
        window
            Window for running average that is applied before plotting.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        self.results.rolling(window=window, center=True).mean().plot(
            ax=ax, **dict(dict(legend=False), **kwargs)
        )
        ax.set(
            title=f"{self.parameter['loc_property']}({self.parameter['index']})\n (window={window})",
            xlabel=self.parameter["index"],
            ylabel=self.parameter["loc_property"],
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: int | Sequence[int | float] | str = "auto",
        log: bool = True,
        fit: bool | None = True,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        hist(results). Nan entries are ignored.

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            Bin specifications (passed to :func:`matplotlib.hist`).
        log
            Flag for plotting on a log scale.
        fit
            Flag indicating if distribution fit is shown.
            The fit will only be computed if `distribution_statistics` is None.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        ax.hist(
            self.results.dropna(axis=0).values,
            bins=bins,
            **dict(dict(density=True, log=log), **kwargs),
        )
        ax.set(
            title=self.parameter["loc_property"],
            xlabel=self.parameter["loc_property"],
            ylabel="PDF",
        )

        # fit distributions:
        if fit:
            if self.distribution_statistics is None:
                self.fit_distributions()

            assert (  # type narrowing # noqa: S101
                self.distribution_statistics is not None
            )
            self.distribution_statistics.plot(ax=ax)

        return ax


# todo add Dependence_stats to fit a plot to a linear function, log function, or exponential decay.


class _DistributionFits:
    """
    Handle for distribution fits.

    This class is typically instantiated by LocalizationProperty methods.
    It holds the statistical parameters derived by fitting the result
    distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationProperty
        The analysis class with result data to fit.
    loc_property : str
        The LocData property for which to fit an appropriate distribution.
    distribution : scipy.stats.rv_continuous | None
        Distribution model to fit.
    parameters : list[str]
        Distribution parameters.
    """

    def __init__(self, analysis_class: LocalizationProperty) -> None:
        self.analysis_class = analysis_class
        self.loc_property = self.analysis_class.parameter["loc_property"]
        self.distribution: stats.rv_continuous | None = None
        self.parameters: list[str] = []

    def fit(
        self,
        distribution: stats.rv_continuous,
        with_constraints: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Fit scipy.stats.rv_continuous to analysis_class.results[loc_property].

        If with_constraints is true we put the following constraints on the
        fit procedure:
        If distribution is expon then
        floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution
            Distribution model to fit.
        with_constraints
            Flag to use predefined constraints on fit parameters.
        kwargs
            Other parameters are passed to `scipy.stats.rv_continuous.fit()`.
        """
        if self.analysis_class.results is None:
            logger.warning("No results available to fit.")
            return None
        self.distribution = distribution
        for param in _list_parameters(distribution):
            self.parameters.append(self.loc_property + "_" + param)

        if with_constraints and self.distribution == stats.expon:
            # MLE fit of exponential distribution with constraints
            fit_results = stats.expon.fit(
                self.analysis_class.results[self.loc_property].values,
                **dict(
                    dict(
                        floc=np.min(
                            self.analysis_class.results[self.loc_property].values
                        )
                    ),
                    **kwargs,
                ),
            )
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)
        else:
            fit_results = self.distribution.fit(
                self.analysis_class.results[self.loc_property].values, **kwargs
            )
            for parameter, result in zip(self.parameters, fit_results):
                setattr(self, parameter, result)

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        probability distribution functions of fitted results.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            The axes on which to show the image.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        parameter = self.parameter_dict().values()
        x_values = np.linspace(
            self.distribution.ppf(0.001, *parameter),
            self.distribution.ppf(0.999, *parameter),
            100,
        )
        ax.plot(
            x_values,
            self.distribution.pdf(x_values, *parameter),
            "r-",
            **dict(
                dict(lw=3, alpha=0.6, label=str(self.distribution.name) + " pdf"),
                **kwargs,
            ),
        )

        return ax

    def parameter_dict(self) -> dict[str, float]:
        """Dictionary of fitted parameters."""
        return {k: self.__dict__[k] for k in self.parameters}



"""
Simulate drift and apply it to localization data.

Drift can be linear in time or resemble a random walk.
"""

from __future__ import annotations

import sys

import numpy as np
import numpy.typing as npt
import pandas as pd

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.locan_types import RandomGeneratorSeed

__all__: list[str] = ["add_drift"]


def _random_walk_drift(
    n_steps: int,
    diffusion_constant: npt.ArrayLike,
    velocity: npt.ArrayLike,
    seed: RandomGeneratorSeed = None,
) -> npt.NDArray[np.float64]:
    """
    Transform locdata coordinates according to a simulated drift.

    Position deltas are computed as function of frame number.
    Within a single time unit delta_t the probability for a diffusion step of
    length delta_x is:

    p(delta_x, delta_t) = Norm(-velocity*delta_t, sigma**2) where the standard
    deviation sigma**2 = 2 * D * delta_t

    Parameters
    ----------
    n_steps
        The number of time steps for the random walk.
    diffusion_constant
        Diffusion constant for each dimension specifying the drift velocity
        with shape (point_dimension,).
        The diffusion constant has the unit of square of localization
        coordinate unit per frame unit.
    velocity
        Drift velocity in units of localization coordinate unit per frame unit
        with shape (point_dimension,)
    seed
        random number generation seed

    Returns
    -------
    npt.NDArray[np.float64]
        Position deltas with shape (n_points, diffusion_constant.shape)
        that have to be added to the original localizations.
    """
    rng = np.random.default_rng(seed)
    diffusion_constant = np.asarray(diffusion_constant)
    velocity = np.asarray(velocity)

    sigmas = np.sqrt(2 * diffusion_constant)  # * delta_t which is 1 frame
    steps = np.array(
        [
            rng.normal(loc=-vel, scale=sigma, size=n_steps)
            for vel, sigma in zip(velocity, sigmas)
        ]
    )
    return np.cumsum(steps, axis=1)


def _drift(
    frames: npt.ArrayLike,
    diffusion_constant: tuple[float, ...] | None = None,
    velocity: tuple[float, ...] | None = None,
    seed: RandomGeneratorSeed = None,
) -> npt.NDArray[np.float64]:
    """
    Compute position deltas as function of frame number.

     Parameters
    ----------
    frames
        array with frame numbers
    diffusion_constant
        Diffusion constant for each dimension specifying the drift velocity
        with shape (point_dimension,).
        The diffusion constant has the unit of square of localization
        coordinate unit per frame unit.
        If None only linear drift is computed.
    velocity
        Drift velocity in units of localization coordinate unit per frame
        unit with shape (point_dimension,)
    seed
        random number generation seed

    Returns
    -------
    npt.NDArray[np.float64]
        Position deltas with shape (n_points, diffusion_constant.shape)
        that have to be added to the original localizations.
    """
    frames_ = np.asarray(frames)

    if diffusion_constant is None and velocity is not None:  # only linear drift
        position_deltas = frames_[:, np.newaxis] * velocity
        position_deltas = position_deltas.T

    elif diffusion_constant is not None:  # random walk plus linear drift
        velocity_ = np.zeros(len(diffusion_constant)) if velocity is None else velocity
        cumsteps = _random_walk_drift(
            frames_.max() + 1, diffusion_constant, velocity_, seed  # type: ignore
        )
        position_deltas = cumsteps[:, frames_]

    else:  # no drift
        position_deltas = None

    return position_deltas


def add_drift(
    locdata: LocData,
    diffusion_constant: tuple[float, ...] | None = None,
    velocity: tuple[float, ...] | None = None,
    seed: RandomGeneratorSeed = None,
) -> LocData:
    """
    Compute position deltas as function of frame number.

    Parameters
    ----------
    locdata
        Original localization data
    diffusion_constant
        Diffusion constant for each dimension specifying the drift velocity
        with shape (point_dimension,).
        The diffusion constant has the unit of square of
        localization coordinate unit per frame unit.
    velocity
        Drift velocity in units of localization coordinate unit per frame
        unit with shape (point_dimension,)
    seed
        random number generation seed

    Returns
    -------
    LocData
        A new LocData instance with localization data.
    """
    local_parameter = locals()

    if len(locdata) == 0:
        return locdata

    points = locdata.coordinates
    frames = locdata.data["frame"]

    if diffusion_constant is None and velocity is None:
        transformed_points = points
    else:
        transformed_points = (
            points + _drift(frames, diffusion_constant, velocity, seed).T
        )

    # new LocData object
    new_dataframe = locdata.data.copy()

    df = pd.DataFrame(
        transformed_points,
        columns=locdata.coordinate_keys,
        index=locdata.data.index,
    )
    # cast dtypes
    df = df.astype(new_dataframe[locdata.coordinate_keys].dtypes)

    new_dataframe.update(df)
    new_locdata = LocData.from_dataframe(new_dataframe)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata



"""

Transform localization data.

This module provides functions that take locdata as input, transform the
localization data, and return a new LocData object.


Submodules:
-----------

.. autosummary::
   :toctree: ./

   bunwarpj
   spatial_transformation
   intensity_transformation

"""

from __future__ import annotations

from locan.process.transform.bunwarpj import *
from locan.process.transform.intensity_transformation import *
from locan.process.transform.spatial_transformation import *

from . import bunwarpj, intensity_transformation, spatial_transformation

__all__: list[str] = []
__all__.extend(bunwarpj.__all__)
__all__.extend(intensity_transformation.__all__)
__all__.extend(spatial_transformation.__all__)



import numpy as np
import pytest

import locan as lc
from locan import FileType, load_decode_file, load_decode_header
from locan.dependencies import HAS_DEPENDENCY

pytestmark = pytest.mark.skipif(not HAS_DEPENDENCY["h5py"], reason="requires h5py.")


def test_load_DECODE_header():
    columns, meta, decode = load_decode_header(
        path=lc.ROOT_DIR / "tests/test_data/decode_dstorm_data.h5"
    )
    # print(columns, meta, decode)
    assert list(meta.keys()) == ["px_size", "xy_unit"]
    assert list(decode.keys()) == ["version"]
    assert columns == [
        "local_background",
        "bg_cr",
        "bg_sig",
        "frame",
        "original_index",
        "intensity",
        "phot_cr",
        "phot_sig",
        "prob",
        "position_x",
        "position_y",
        "position_z",
        "x_cr",
        "y_cr",
        "z_cr",
        "x_sig",
        "y_sig",
        "z_sig",
    ]


def test_loading_DECODE_file_empty_file():
    locdata = load_decode_file(
        path=lc.ROOT_DIR / "tests/test_data/decode_dstorm_data_empty.h5", nrows=10
    )
    assert len(locdata) == 0


def test_loading_DECODE_file():
    file_path = lc.ROOT_DIR / "tests/test_data/decode_dstorm_data.h5"
    locdata = load_decode_file(path=file_path, nrows=10)
    # print(locdata.data)
    assert len(locdata) == 10
    assert np.array_equal(
        locdata.data.columns,
        [
            "local_background",
            "frame",
            "original_index",
            "intensity",
            "phot_sig",
            "prob",
            "position_x",
            "position_y",
            "position_z",
            "x_sig",
            "y_sig",
            "z_sig",
        ],
    )
    assert locdata.meta.file.type == FileType.DECODE.value
    assert locdata.meta.file.path == str(file_path)



"""

Utility functions for working with locdata.

"""

from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from locan.utils.statistics import weighted_mean_variance

if TYPE_CHECKING:
    from locan.data.locdata import LocData  # noqa F401
    from locan.locan_types import DataFrame  # noqa F401

__all__: list[str] = []

logger = logging.getLogger(__name__)


def _get_loc_property_key_per_dimension(
    locdata: pd.DataFrame | pd.Series[Any], property_key: str
) -> list[str | None]:
    """
    Get tuple with property_key as available in each dimension.
    For the x-coordinates precedence is given as:
    `position_x` or `position` or None.
    None if the key is not available for a certain dimension.

    Parameters
    ----------
    locdata
        Localization data
    property_key
        Property key to look for in locdata.

    Returns
    -------
    list[str | None]
        The available property keys in each dimension
    """
    available_keys: list[str | None] = []
    for extension_ in ["_x", "_y", "_z"]:
        prop_key = property_key + extension_
        if prop_key in locdata.columns:
            available_keys.append(prop_key)
        elif property_key in locdata.columns:
            available_keys.append(property_key)
        else:
            available_keys.append(None)
    return available_keys


def _get_linked_coordinates(
    locdata: pd.DataFrame | pd.Series[Any], coordinate_keys: Iterable[str] | None = None
) -> dict[str, int | float]:
    """
    Combine localization properties from locdata:
    (i) apply weighted averages for spatial coordinates if corresponding
    uncertainties are available;
    (ii) compute uncertainty (as standard deviation with same units as
    coordinates) from weighted average variances for spatial coordinates.

    If coordinate_keys is None, coordinate_keys and corresponding
    coordinate uncertainties from locdata are used.

    If `uncertainty_x` is not available `uncertainty` is taken.
    If no uncertainty is available, unweighted coordinate means are taken.

    Note
    -----
    Uncertainties should have the same unit as coordinates;
    the weights will be 1 / uncertainties^2.

    Parameters
    ----------
    locdata
        dataframe with locdata
    coordinate_keys
        A selection of coordinate keys on which to compute

    Returns
    -------
    dict[str, int | float]
        New position coordinates and related uncertainties as 'uncertainty_c'.
    """
    available_coordinate_keys = _get_loc_property_key_per_dimension(
        locdata=locdata, property_key="position"
    )
    available_uncertainty_keys = _get_loc_property_key_per_dimension(
        locdata=locdata, property_key="uncertainty"
    )
    new_uncertainty_keys = ["uncertainty_x", "uncertainty_y", "uncertainty_z"]

    results_dict = dict()
    for coordinate_key_, uncertainty_key_, new_uncertainty_key_ in zip(
        available_coordinate_keys,
        available_uncertainty_keys,
        new_uncertainty_keys,
    ):
        if coordinate_key_ is not None and (
            coordinate_keys is None or coordinate_key_ in coordinate_keys
        ):
            if not len(locdata):
                return dict()
            elif len(locdata) == 1:
                weighted_mean = locdata[coordinate_key_].iloc[0]
                if uncertainty_key_ is None:
                    weighted_uncertainty = 0
                else:
                    weighted_uncertainty = locdata[uncertainty_key_].iloc[0]
            else:
                if uncertainty_key_ is None:
                    weighted_mean, weighted_variance = weighted_mean_variance(
                        values=locdata[coordinate_key_], weights=None
                    )
                    weighted_uncertainty = np.sqrt(weighted_variance)  # type: ignore
                else:
                    if (locdata[uncertainty_key_] == 0).any():
                        logger.warning(
                            "Zero uncertainties occurred resulting in nan for weighted_mean and weighted_variance."
                        )
                    with np.errstate(invalid="ignore"):
                        weighted_mean, weighted_variance = weighted_mean_variance(
                            values=locdata[coordinate_key_],
                            weights=np.power(1 / locdata[uncertainty_key_], 2),
                        )
                        weighted_uncertainty = np.sqrt(weighted_variance)  # type: ignore

            results_dict[coordinate_key_] = weighted_mean
            results_dict[new_uncertainty_key_] = weighted_uncertainty

    return results_dict


def _bump_property_key(
    loc_property: str, loc_properties: Iterable[str], extension: str = "_0"
) -> str:
    """
    Add extension to loc_property if loc_property is in loc_properties.
    Repeat recursively until loc_property and all loc_properties are unique.
    """
    if loc_property in loc_properties:
        new_property_label = _bump_property_key(
            loc_property=loc_property + extension,
            loc_properties=loc_properties,
            extension=extension,
        )
    else:
        new_property_label = loc_property
    return new_property_label


def _check_loc_properties(
    locdata: LocData, loc_properties: str | Iterable[str] | None
) -> list[str]:
    """
    Check that loc_properties are valid property keys in locdata.

    Parameters
    ----------
    locdata
        Localization data
    loc_properties
        LocData property keys.
        If None the coordinate_keys of locdata are used.

    Returns
    -------
    list[str]
        Valid localization property keys
    """
    if loc_properties is None:  # use coordinate_keys
        labels = locdata.coordinate_keys.copy()
    elif isinstance(loc_properties, str):
        if loc_properties not in locdata.data.columns:
            raise ValueError(
                f"{loc_properties} is not a valid property in locdata.data."
            )
        labels = [loc_properties]
    else:
        labels = []
        for loc_property in loc_properties:
            if loc_property not in locdata.data.columns:
                raise ValueError(
                    f"{loc_property} is not a valid property in locdata.data."
                )
            labels.append(loc_property)
    return labels


def _dataframe_to_pandas(
    dataframe: DataFrame | None, allow_copy: bool = True
) -> pd.DataFrame | None:
    """
    Convert dataframe that supports the dataframe interchange protocol to
    pandas dataframe.
    """
    if isinstance(dataframe, pd.DataFrame) or dataframe is None:
        return_value = dataframe
    elif not hasattr(pd.api, "interchange"):
        msg = (
            "Use pandas version that implements the DataFrame interchange protocol"
            "or provide pandas.DataFrame object."
        )
        raise TypeError(msg)
    else:
        return_value = pd.api.interchange.from_dataframe(
            dataframe, allow_copy=allow_copy
        )
    return return_value



"""

This module provides functions for rendering `LocData` objects in 3D.

"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from matplotlib import pyplot as plt

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data import LocData

__all__: list[str] = ["scatter_3d_mpl"]

logger = logging.getLogger(__name__)


def scatter_3d_mpl(
    locdata: LocData,
    ax: mpl.axes.Axes | None = None,
    index: bool = True,
    text_kwargs: dict[str, Any] | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Scatter plot of locdata elements with text marker for each element.

    Parameters
    ----------
    locdata
       Localization data.
    ax
       The axes on which to show the plot
    index
       Flag indicating if element indices are shown.
    text_kwargs
       Keyword arguments for :func:`matplotlib.axes.Axes.text`.
    kwargs
       Other parameters passed to :func:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    matplotlib.axes.Axes3D
       Axes object with the image.
    """
    if text_kwargs is None:
        text_kwargs = {}

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax

    coordinates = locdata.coordinates
    ax.scatter(*coordinates.T, **dict({"marker": "+", "color": "grey"}, **kwargs))

    # plot element number
    if index:
        for centroid, marker in zip(coordinates, locdata.data.index.values):
            ax.text(  # type:ignore
                *centroid, marker, **dict({"color": "grey", "size": 20}, **text_kwargs)
            )

    ax.set(xlabel="position_x", ylabel="position_y", zlabel="position_z")

    return ax



"""
Benchmark functions for :mod:`locan.analysis.localization_precision`
"""

import locan as lc
from locan.analysis.localization_precision import (
    _localization_precision,
)


class BenchmarkLocalizationPrecision:
    """
    Benchmarks for Ripley's k function computations
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)
        print(self.locdata.data.info())

    def check__localization_precision(self):
        results_df = _localization_precision(locdata=self.locdata)
        print(results_df)

    def time__localization_precision(self):
        _localization_precision(locdata=self.locdata)


def main():
    bm = BenchmarkLocalizationPrecision()
    bm.setup()
    bm.check__localization_precision()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkLocalizationPrecision()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(3):
        bm.time__localization_precision()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    main()
    main_profile()



"""
Utility script to print colormap colors as list.

Such color lists are copied into locan.visualize.colormap_definitions for
default colormaps.
"""

import numpy as np
from colorcet import m_coolwarm, m_fire, m_glasbey, m_glasbey_dark, m_gray  # noqa: F401

from locan.visualize.colormap import ColormapType


def get_colors(colormap: ColormapType) -> None:
    """
    Extract RGB color values from colormap and print as list of lists.

    Parameters
    ----------
    colormap
        The colormap to be evaluated.

    Returns
    -------
    None
    """
    print(colormap.name)
    curated_list = colormap(np.linspace(0, 1, 256))
    print("[")
    for item in curated_list:
        print("[" + ", ".join(map(str, item)) + "],")
    print("]")


if __name__ == "__main__":
    get_colors(colormap=m_glasbey_dark)



"""

File input/output for localization data in Elyra files.

"""

from __future__ import annotations

import io
import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsReadline

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
    open_path_or_file_like,
)

__all__: list[str] = ["load_Elyra_header", "load_Elyra_file"]

logger = logging.getLogger(__name__)


def _read_Elyra_header(file: SupportsReadline[Any]) -> list[str]:
    """
    Read xml header from a Zeiss Elyra single-molecule localization file and
    identify column names.

    Parameters
    ----------
    file
        A file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """
    header = file.readline().split("\n")[0]

    # list identifiers
    identifiers = header.split("\t")

    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.ELYRA_KEYS
    )

    return column_keys


def load_Elyra_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> list[str]:
    """
    Load xml header from a Zeiss Elyra single-molecule localization file and
    identify column names.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """

    with open_path_or_file_like(path, encoding="latin-1") as file:
        return _read_Elyra_header(file)


def load_Elyra_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
    **kwargs: Any,
) -> LocData:
    """
    Load data from a rapidSTORM single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.

    Note
    ----
    Data is loaded with encoding = 'latin-1' and only data before the first
    NUL character is returned.
    Additional information appended at the end of the file is thus ignored.
    """
    with open_path_or_file_like(path, encoding="latin-1") as file:
        columns = _read_Elyra_header(file)
        string = file.read()
        # remove metadata following nul byte
        string = string.split("\x00")[0]

        stream = io.StringIO(string)
        dataframe = pd.read_csv(
            stream, sep="\t", skiprows=0, nrows=nrows, names=columns, **kwargs
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.ELYRA
    dat.meta.file.path = str(path)

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_Elyra_file",
        parameter=f"path={str(path)}, nrows={nrows}",
    )

    return dat



"""
Coordinate-based colocalization.

Colocalization is estimated by computing a colocalization index for each
localization using the so-called coordinate-based colocalization
algorithm [1]_.

References
----------
.. [1] Malkusch S, Endesfelder U, Mondry J, Gelléri M, Verveer PJ, Heilemann M.,
   Coordinate-based colocalization analysis of single-molecule localization microscopy data.
   Histochem Cell Biol. 2012, 137(1):1-10.
   doi: 10.1007/s00418-011-0880-5
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.stats import spearmanr
from sklearn.neighbors import NearestNeighbors

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis

__all__: list[str] = ["CoordinateBasedColocalization"]

logger = logging.getLogger(__name__)


# The algorithm


def _coordinate_based_colocalization(
    points: npt.ArrayLike,
    other_points: npt.ArrayLike | None = None,
    radius: int | float = 100,
    n_steps: int = 10,
) -> npt.NDArray[np.float64]:
    """
    Compute a colocalization index for each localization by coordinate-based
    colocalization.

    Parameters
    ----------
    points
        Array of points with shape (n_points, dimensions)
        for which CBC values are computed.

    other_points
        Array of points with shape (n_points, dimensions) to be
        compared with points. If None other_points are set to points.

    radius
        The maximum radius up to which nearest neighbors are determined

    n_steps
        The number of bins from which Spearman correlation is computed.

    Returns
    -------
    npt.NDArray[np.float64]
        An array with coordinate-based colocalization coefficients for each
        input point.
    """
    points = np.asarray(points)
    # sampled radii
    radii = np.linspace(0, radius, n_steps + 1)

    # nearest neighbors within radius
    nneigh_1 = NearestNeighbors(radius=radius, metric="euclidean").fit(points)
    distances_1 = np.array(nneigh_1.radius_neighbors()[0])

    if other_points is None:
        nneigh_2 = NearestNeighbors(radius=radius, metric="euclidean").fit(points)
    else:
        nneigh_2 = NearestNeighbors(radius=radius, metric="euclidean").fit(other_points)
    distances_2 = np.array(nneigh_2.radius_neighbors(points)[0])

    # CBC for each point
    correlation = np.empty(len(points))
    for i, (d_1, d_2) in enumerate(zip(distances_1, distances_2)):
        if len(d_1) and len(d_2):
            # binning
            hist_1 = np.histogram(d_1, bins=radii, range=(0, radius))[0]
            hist_2 = np.histogram(d_2, bins=radii, range=(0, radius))[0]

            # normalization
            values_1 = np.cumsum(hist_1) * radius**2 / radii[1:] ** 2 / len(d_1)
            values_2 = np.cumsum(hist_2) * radius**2 / radii[1:] ** 2 / len(d_2)

            # Spearman rank correlation
            rho, _pval = spearmanr(values_1, values_2)
            correlation[i] = rho
        else:
            correlation[i] = np.nan

    # CBC normalization for each point
    max_distances = np.array(
        [np.max(d, initial=0) for d in distances_2]
    )  # max is set to 0 for empty arrays.
    norm_spearmanr = np.exp(-1 * max_distances / radius)
    correlation = correlation * norm_spearmanr

    return correlation


# The specific analysis classes


class CoordinateBasedColocalization(_Analysis):
    """
    Compute a colocalization index for each localization by coordinate-based
    colocalization (CBC).

    The colocalization index is calculated for each localization in `locdata`
    by finding nearest neighbors in `locdata` or `other_locdata` within
    `radius`. A normalized number of nearest neighbors at a certain radius is
    computed for `n_steps` equally-sized steps of increasing radii ranging
    from 0 to `radius`.
    The Spearman rank correlation coefficent is computed for these values and
    weighted by Exp[-nearestNeighborDistance/distanceMax].

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    radius : int | float
        The maximum radius up to which nearest neighbors are determined
    n_steps : int
        The number of bins from which Spearman correlation is computed.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Coordinate-based colocalization coefficients for each input point.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        radius: int | float = 100,
        n_steps: int = 10,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData, other_locdata: LocData | None = None) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data for which CBC values are computed.
        other_locdata
            Localization data to be colocalized.
            If None other_locdata is set to locdata.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
            id_ = other_locdata.meta.identifier
        else:
            other_points = None
            id_ = "self"

        self.results = pd.DataFrame(
            {
                f"colocalization_cbc_{id_}": _coordinate_based_colocalization(
                    points, other_points, **self.parameter
                )
            }
        )
        return self

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: int | Sequence[int | float] | Literal["auto"] = (-1, -0.3333, 0.3333, 1),
        density: bool = True,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        hist(results).

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            Bin specification as used in :func:`matplotlib.hist`.
        density
            Flag for normalization as used in :func:`matplotlib.hist`.
            True returns probability density function;
            None returns counts.
        kwargs
            Other parameters passed to :func:`matplotlib.hist`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        ax.hist(
            self.results.iloc[:, 0].values,
            bins=bins,
            density=density,
            label="cbc",
            **kwargs,
        )

        ax.set(
            title="CBC histogram",
            xlabel="colocalization_cbc",
            ylabel="pdf" if density else "counts",
        )
        ax.legend(loc="best")

        return ax



from pathlib import Path

from locan.configuration import (
    DATASETS_DIR,
    N_JOBS,
    QT_BINDING,
    RENDER_ENGINE,
)
from locan.constants import RenderEngine
from locan.dependencies import HAS_DEPENDENCY


def test_datasets_dir():
    # print(DATASETS_DIR)
    assert isinstance(DATASETS_DIR, Path)


def test_render_engine():
    assert isinstance(RENDER_ENGINE, RenderEngine)


def test_n_jobs():
    assert isinstance(N_JOBS, int)


def test_qt_bindings():
    assert "qt" in HAS_DEPENDENCY
    assert isinstance(QT_BINDING, str)

    if HAS_DEPENDENCY["qt"]:
        from qtpy import QT_VERSION

        assert QT_VERSION is not None


# tests for COLORMAP_DEFAULTS see test_colormap.py



"""

Configuration variables used throughout the project.

.. autosummary::
   :toctree: ./

   DATASETS_DIR
   RENDER_ENGINE
   N_JOBS
   COLORMAP_DEFAULTS
   TQDM_LEAVE
   TQDM_DISABLE
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from pathlib import Path

from locan.constants import RenderEngine
from locan.dependencies import HAS_DEPENDENCY, QtBindings, _set_qt_binding

logger = logging.getLogger(__name__)

__all__: list[str] = [
    "DATASETS_DIR",
    "RENDER_ENGINE",
    "N_JOBS",
    "TQDM_LEAVE",
    "TQDM_DISABLE",
    "COLORMAP_DEFAULTS",
    "QT_BINDING",
]


#: Standard directory for example datasets.
DATASETS_DIR = Path.home() / "LocanDatasets"

#: Render engine.
RENDER_ENGINE = RenderEngine.MPL

#: The number of cores that are used in parallel for some algorithms.
#: Following the scikit convention: n_jobs is the number of parallel jobs to run.
#: If -1, then the number of jobs is set to the number of CPU cores.
N_JOBS: int = 1

#: Leave tqdm progress bars after finishing the iteration.
#: Flag to leave tqdm progress bars.
TQDM_LEAVE: bool = True

#: Disable tqdm progress bars.
#: Flag to disable all tqdm progress bars.
TQDM_DISABLE: bool = False


#: Mapping a type of colormap to its default colormap that is used throughout locan.
#: See details in locan documentation on colormaps.
COLORMAP_DEFAULTS: Mapping[str, str] = dict()

COLORMAP_DEFAULTS = dict(
    CONTINUOUS="cet_fire",
    CONTINUOUS_REVERSE="cet_fire_r",
    CONTINUOUS_GRAY="cet_gray",
    CONTINUOUS_GRAY_REVERSE="cet_gray_r",
    DIVERGING="cet_coolwarm",
    CATEGORICAL="cet_glasbey_dark",
    TURBO="turbo",
)


# Preferred QT binding. At this point this variable cannot be set at
# runtime.
QT_BINDING: QtBindings | str = ""

# Initialize pyqt
QT_BINDING = _set_qt_binding(QT_BINDING)
HAS_DEPENDENCY["qt"] = True if QT_BINDING else False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: locan/locan_io/locdata/manifest.proto
# Protobuf Python Version: 5.28.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    2,
    "",
    "locan/locan_io/locdata/manifest.proto",
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from locan.data import metadata_pb2 as locan_dot_data_dot_metadata__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%locan/locan_io/locdata/manifest.proto\x12\x16locan.locan_io.locdata\x1a\x19locan/data/metadata.proto"\xf1\x01\n\x06\x46ormat\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12*\n\x04mode\x18\x03 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Mode\x12\x0f\n\x07\x63olumns\x18\x04 \x01(\x03\x12\x0f\n\x07headers\x18\x05 \x03(\t\x12,\n\x05\x64type\x18\x06 \x03(\x0e\x32\x1d.locan.locan_io.locdata.Dtype\x12\r\n\x05shape\x18\x07 \x03(\x03\x12\r\n\x05units\x18\x08 \x03(\t\x12\x13\n\x0b\x64\x65scription\x18\t \x01(\t"\xba\x03\n\x08\x46ileInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04type\x18\x02 \x01(\x0e\x32\x1c.locan.locan_io.locdata.Type\x12\x0e\n\x06\x66ormat\x18\x03 \x01(\t\x12\x0f\n\x07\x63hannel\x18\x04 \x01(\t\x12\x0c\n\x04rows\x18\x05 \x01(\x03\x12<\n\x06offset\x18\x06 \x03(\x0b\x32,.locan.locan_io.locdata.FileInfo.OffsetEntry\x12\x36\n\x03min\x18\x07 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MinEntry\x12\x36\n\x03max\x18\x08 \x03(\x0b\x32).locan.locan_io.locdata.FileInfo.MaxEntry\x12\x10\n\x08\x65xposure\x18\t \x01(\x02\x1a-\n\x0bOffsetEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a*\n\x08MinEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a*\n\x08MaxEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01"\x98\x04\n\x08Manifest\x12\x16\n\x0e\x66ormat_version\x18\x01 \x01(\t\x12>\n\x07\x66ormats\x18\x02 \x03(\x0b\x32-.locan.locan_io.locdata.Manifest.FormatsEntry\x12/\n\x05\x66iles\x18\x03 \x03(\x0b\x32 .locan.locan_io.locdata.FileInfo\x12\x0c\n\x04name\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x11\n\tthumbnail\x18\x07 \x01(\t\x12\x0e\n\x06sample\x18\x08 \x01(\t\x12\x10\n\x08labeling\x18\t \x01(\t\x12\x0c\n\x04\x64\x61te\x18\n \x01(\t\x12\x0e\n\x06\x61uthor\x18\x0b \x01(\t\x12\x10\n\x08\x63itation\x18\x0c \x01(\t\x12\r\n\x05\x65mail\x18\r \x01(\t\x12\x36\n\x03map\x18\x0e \x03(\x0b\x32).locan.locan_io.locdata.Manifest.MapEntry\x12*\n\x0clocdata_meta\x18\x0f \x01(\x0b\x32\x14.locan.data.Metadata\x1aN\n\x0c\x46ormatsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x1e.locan.locan_io.locdata.Format:\x02\x38\x01\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01*\x1c\n\x04Type\x12\t\n\x05TABLE\x10\x00\x12\t\n\x05IMAGE\x10\x01*\x1c\n\x04Mode\x12\n\n\x06\x42INARY\x10\x00\x12\x08\n\x04TEXT\x10\x01*{\n\x05\x44type\x12\x08\n\x04INT8\x10\x01\x12\t\n\x05UINT8\x10\x02\x12\t\n\x05INT16\x10\x03\x12\n\n\x06UINT16\x10\x04\x12\t\n\x05INT32\x10\x05\x12\n\n\x06UINT32\x10\x06\x12\t\n\x05INT64\x10\x07\x12\n\n\x06UINT64\x10\x08\x12\x0b\n\x07\x46LOAT32\x10\t\x12\x0b\n\x07\x46LOAT64\x10\n'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "locan.locan_io.locdata.manifest_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_FILEINFO_OFFSETENTRY"]._loaded_options = None
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_options = b"8\001"
    _globals["_FILEINFO_MINENTRY"]._loaded_options = None
    _globals["_FILEINFO_MINENTRY"]._serialized_options = b"8\001"
    _globals["_FILEINFO_MAXENTRY"]._loaded_options = None
    _globals["_FILEINFO_MAXENTRY"]._serialized_options = b"8\001"
    _globals["_MANIFEST_FORMATSENTRY"]._loaded_options = None
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_options = b"8\001"
    _globals["_MANIFEST_MAPENTRY"]._loaded_options = None
    _globals["_MANIFEST_MAPENTRY"]._serialized_options = b"8\001"
    _globals["_TYPE"]._serialized_start = 1320
    _globals["_TYPE"]._serialized_end = 1348
    _globals["_MODE"]._serialized_start = 1350
    _globals["_MODE"]._serialized_end = 1378
    _globals["_DTYPE"]._serialized_start = 1380
    _globals["_DTYPE"]._serialized_end = 1503
    _globals["_FORMAT"]._serialized_start = 93
    _globals["_FORMAT"]._serialized_end = 334
    _globals["_FILEINFO"]._serialized_start = 337
    _globals["_FILEINFO"]._serialized_end = 779
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_start = 646
    _globals["_FILEINFO_OFFSETENTRY"]._serialized_end = 691
    _globals["_FILEINFO_MINENTRY"]._serialized_start = 693
    _globals["_FILEINFO_MINENTRY"]._serialized_end = 735
    _globals["_FILEINFO_MAXENTRY"]._serialized_start = 737
    _globals["_FILEINFO_MAXENTRY"]._serialized_end = 779
    _globals["_MANIFEST"]._serialized_start = 782
    _globals["_MANIFEST"]._serialized_end = 1318
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_start = 1196
    _globals["_MANIFEST_FORMATSENTRY"]._serialized_end = 1274
    _globals["_MANIFEST_MAPENTRY"]._serialized_start = 1276
    _globals["_MANIFEST_MAPENTRY"]._serialized_end = 1318
# @@protoc_insertion_point(module_scope)
